from functools import partial
from pathlib import Path

import hydra
import mlflow
import numpy as np
import torch
import torch.distributed as dist
from hydra.utils import to_absolute_path
from nnmnkwii import metrics
from nnsvs.base import PredictionType
from nnsvs.mdn import mdn_get_most_probable_sigma_and_mu, mdn_loss
from nnsvs.multistream import split_streams
from nnsvs.train_util import (
    collate_fn_default,
    collate_fn_random_segments,
    get_stream_weight,
    log_params_from_omegaconf_dict,
    save_checkpoint,
    save_configs,
    setup,
)
from nnsvs.util import PyTorchStandardScaler, make_non_pad_mask
from omegaconf import DictConfig
from torch import nn
from torch.cuda.amp import autocast


@torch.no_grad()
def compute_distortions(pred_out_feats, out_feats, lengths, out_scaler):
    assert pred_out_feats.shape == out_feats.shape
    out_feats = out_scaler.inverse_transform(out_feats)
    pred_out_feats = out_scaler.inverse_transform(pred_out_feats)

    dist = {}
    try:
        dist["ObjEval_RMSE"] = np.sqrt(
            metrics.mean_squared_error(out_feats, pred_out_feats, lengths=lengths)
        )
    except ZeroDivisionError:
        pass

    return dist


def train_step(
    model,
    optimizer,
    grad_scaler,
    train,
    in_feats,
    out_feats,
    lengths,
    out_scaler,
    feats_criterion="mse",
    stream_wise_loss=False,
    stream_weights=None,
    stream_sizes=None,
):
    model.train() if train else model.eval()
    optimizer.zero_grad()

    if feats_criterion in ["l2", "mse"]:
        criterion = nn.MSELoss(reduction="none")
    elif feats_criterion in ["l1", "mae"]:
        criterion = nn.L1Loss(reduction="none")
    else:
        raise RuntimeError("not supported criterion")

    prediction_type = (
        model.module.prediction_type()
        if isinstance(model, nn.DataParallel)
        else model.prediction_type()
    )

    # Apply preprocess if required (e.g., FIR filter for shallow AR)
    # defaults to no-op
    if isinstance(model, nn.DataParallel):
        out_feats = model.module.preprocess_target(out_feats)
    else:
        out_feats = model.preprocess_target(out_feats)

    # Run forward
    with autocast(enabled=grad_scaler is not None):
        pred_out_feats = model(in_feats, lengths)

    # Mask (B, T, 1)
    mask = make_non_pad_mask(lengths).unsqueeze(-1).to(in_feats.device)

    # Compute loss
    if prediction_type == PredictionType.PROBABILISTIC:
        pi, sigma, mu = pred_out_feats
        # (B, max(T)) or (B, max(T), D_out)
        mask_ = mask if len(pi.shape) == 4 else mask.squeeze(-1)
        # Compute loss and apply mask
        with autocast(enabled=grad_scaler is not None):
            loss = mdn_loss(pi, sigma, mu, out_feats, reduce=False)
        loss = loss.masked_select(mask_).mean()
    else:
        if stream_wise_loss:
            w = get_stream_weight(stream_weights, stream_sizes).to(in_feats.device)
            streams = split_streams(out_feats, stream_sizes)
            pred_streams = split_streams(pred_out_feats, stream_sizes)
            loss = 0
            for pred_stream, stream, sw in zip(pred_streams, streams, w):
                with autocast(enabled=grad_scaler is not None):
                    loss += (
                        sw
                        * criterion(
                            pred_stream.masked_select(mask), stream.masked_select(mask)
                        ).mean()
                    )
        else:
            with autocast(enabled=grad_scaler is not None):
                loss = criterion(
                    pred_out_feats.masked_select(mask), out_feats.masked_select(mask)
                ).mean()

    if prediction_type == PredictionType.PROBABILISTIC:
        with torch.no_grad():
            pred_out_feats_ = mdn_get_most_probable_sigma_and_mu(pi, sigma, mu)[1]
    else:
        pred_out_feats_ = pred_out_feats
    distortions = compute_distortions(pred_out_feats_, out_feats, lengths, out_scaler)

    if train:
        if grad_scaler is not None:
            grad_scaler.scale(loss).backward()
            grad_scaler.step(optimizer)
            grad_scaler.update()
        else:
            loss.backward()
            optimizer.step()

    return loss, distortions


def train_loop(
    config,
    logger,
    device,
    model,
    optimizer,
    lr_scheduler,
    grad_scaler,
    data_loaders,
    samplers,
    writer,
    out_scaler,
    use_mlflow,
):
    out_dir = Path(to_absolute_path(config.train.out_dir))
    best_dev_loss = torch.finfo(torch.float32).max
    last_dev_loss = torch.finfo(torch.float32).max

    if "feats_criterion" not in config.train:
        logger.warning(
            "feats_criterion is not found in the train config. Fallback to MSE."
        )
        feats_criterion = "mse"
    else:
        feats_criterion = config.train.feats_criterion

    if dist.is_initialized() and dist.get_rank() != 0:

        def tqdm(x, **kwargs):
            return x

    else:
        from tqdm import tqdm

    train_iter = 1
    for epoch in tqdm(range(1, config.train.nepochs + 1)):
        for phase in data_loaders.keys():
            train = phase.startswith("train")
            # https://pytorch.org/docs/stable/data.html#torch.utils.data.distributed.DistributedSampler
            if dist.is_initialized() and train and samplers[phase] is not None:
                samplers[phase].set_epoch(epoch)
            running_loss = 0
            running_metrics = {}
            for in_feats, out_feats, lengths in tqdm(
                data_loaders[phase], desc=f"{phase} iter", leave=False
            ):
                # NOTE: This is needed for pytorch's PackedSequence
                lengths, indices = torch.sort(lengths, dim=0, descending=True)
                in_feats, out_feats = (
                    in_feats[indices].to(device),
                    out_feats[indices].to(device),
                )
                loss, log_metrics = train_step(
                    model=model,
                    optimizer=optimizer,
                    grad_scaler=grad_scaler,
                    train=train,
                    in_feats=in_feats,
                    out_feats=out_feats,
                    lengths=lengths,
                    out_scaler=out_scaler,
                    feats_criterion=feats_criterion,
                    stream_wise_loss=config.train.stream_wise_loss,
                    stream_weights=config.model.stream_weights,
                    stream_sizes=config.model.stream_sizes,
                )

                if train:
                    if writer is not None:
                        for key, val in log_metrics.items():
                            writer.add_scalar(f"{key}_Step/{phase}", val, train_iter)
                    train_iter += 1

                running_loss += loss.item()
                for k, v in log_metrics.items():
                    try:
                        running_metrics[k] += float(v)
                    except KeyError:
                        running_metrics[k] = float(v)

            ave_loss = running_loss / len(data_loaders[phase])
            if writer is not None:
                writer.add_scalar(f"Loss/{phase}", ave_loss, epoch)
            if use_mlflow:
                mlflow.log_metric(f"{phase}_loss", ave_loss, step=epoch)

            ave_loss = running_loss / len(data_loaders[phase])
            logger.info("[%s] [Epoch %s]: loss %s", phase, epoch, ave_loss)
            if writer is not None:
                writer.add_scalar(f"Loss_Epoch/{phase}", ave_loss, epoch)
            if use_mlflow:
                mlflow.log_metric(f"{phase}_loss", ave_loss, step=epoch)

            for k, v in running_metrics.items():
                ave_v = v / len(data_loaders[phase])
                if writer is not None:
                    writer.add_scalar(f"{k}_Epoch/{phase}", ave_v, epoch)
                if use_mlflow:
                    mlflow.log_metric(f"{phase}_{k}", ave_v, step=epoch)

            if not train:
                last_dev_loss = ave_loss
            if not train and ave_loss < best_dev_loss:
                best_dev_loss = ave_loss
                save_checkpoint(
                    logger, out_dir, model, optimizer, lr_scheduler, epoch, is_best=True
                )

        lr_scheduler.step()
        if epoch % config.train.checkpoint_epoch_interval == 0:
            save_checkpoint(
                logger, out_dir, model, optimizer, lr_scheduler, epoch, is_best=False
            )

    save_checkpoint(
        logger, out_dir, model, optimizer, lr_scheduler, config.train.nepochs
    )
    logger.info("The best loss was %s", best_dev_loss)
    if use_mlflow:
        mlflow.log_metric("best_dev_loss", best_dev_loss, step=epoch)
        mlflow.log_artifacts(out_dir)

    return last_dev_loss


@hydra.main(config_path="conf/train", config_name="config")
def my_app(config: DictConfig) -> None:
    if "max_time_frames" in config.data and config.data.max_time_frames > 0:
        collate_fn = partial(
            collate_fn_random_segments, max_time_frames=config.data.max_time_frames
        )
    else:
        collate_fn = collate_fn_default

    if config.train.use_ddp:
        dist.init_process_group("nccl")
        rank = dist.get_rank()
        device_id = rank % torch.cuda.device_count()
        torch.cuda.set_device(device_id)

    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    (
        model,
        optimizer,
        lr_scheduler,
        grad_scaler,
        data_loaders,
        samplers,
        writer,
        logger,
        _,
        out_scaler,
    ) = setup(config, device, collate_fn)

    out_scaler = PyTorchStandardScaler(
        torch.from_numpy(out_scaler.mean_), torch.from_numpy(out_scaler.scale_)
    ).to(device)
    use_mlflow = config.mlflow.enabled

    if use_mlflow:
        with mlflow.start_run() as run:
            # NOTE: modify out_dir when running with mlflow
            config.train.out_dir = f"{config.train.out_dir}/{run.info.run_id}"
            save_configs(config)
            log_params_from_omegaconf_dict(config)
            last_dev_loss = train_loop(
                config=config,
                logger=logger,
                device=device,
                model=model,
                optimizer=optimizer,
                lr_scheduler=lr_scheduler,
                grad_scaler=grad_scaler,
                data_loaders=data_loaders,
                samplers=samplers,
                writer=writer,
                out_scaler=out_scaler,
                use_mlflow=use_mlflow,
            )
    else:
        save_configs(config)
        last_dev_loss = train_loop(
            config=config,
            logger=logger,
            device=device,
            model=model,
            optimizer=optimizer,
            lr_scheduler=lr_scheduler,
            grad_scaler=grad_scaler,
            data_loaders=data_loaders,
            samplers=samplers,
            writer=writer,
            out_scaler=out_scaler,
            use_mlflow=use_mlflow,
        )

    return last_dev_loss


def entry():
    my_app()


if __name__ == "__main__":
    my_app()



# -*- coding: utf-8 -*-

"""Upsampling module.

This code is modified from https://github.com/r9y9/wavenet_vocoder.

"""

import numpy as np
import torch
import torch.nn.functional as F
from nnsvs.usfgan.layers import Conv1d


class Stretch2d(torch.nn.Module):
    """Stretch2d module."""

    def __init__(self, x_scale, y_scale, mode="nearest"):
        """Initialize Stretch2d module.

        Args:
            x_scale (int): X scaling factor (Time axis in spectrogram).
            y_scale (int): Y scaling factor (Frequency axis in spectrogram).
            mode (str): Interpolation mode.

        """
        super(Stretch2d, self).__init__()
        self.x_scale = x_scale
        self.y_scale = y_scale
        self.mode = mode

    def forward(self, x):
        """Calculate forward propagation.

        Args:
            x (Tensor): Input tensor (B, C, F, T).

        Returns:
            Tensor: Interpolated tensor (B, C, F * y_scale, T * x_scale)

        """
        return F.interpolate(
            x, scale_factor=(self.y_scale, self.x_scale), mode=self.mode
        )


class Conv2d(torch.nn.Conv2d):
    """Conv2d module with customized initialization."""

    def __init__(self, *args, **kwargs):
        """Initialize Conv2d module."""
        super(Conv2d, self).__init__(*args, **kwargs)

    def reset_parameters(self):
        """Reset parameters."""
        self.weight.data.fill_(1.0 / np.prod(self.kernel_size))
        if self.bias is not None:
            torch.nn.init.constant_(self.bias, 0.0)


class UpsampleNetwork(torch.nn.Module):
    """Upsampling network module."""

    def __init__(
        self,
        upsample_scales,
        nonlinear_activation=None,
        nonlinear_activation_params={},
        interpolate_mode="nearest",
        freq_axis_kernel_size=1,
        use_causal_conv=False,
    ):
        """Initialize upsampling network module.

        Args:
            upsample_scales (list): List of upsampling scales.
            nonlinear_activation (str): Activation function name.
            nonlinear_activation_params (dict): Arguments for specified activation function.
            interpolate_mode (str): Interpolation mode.
            freq_axis_kernel_size (int): Kernel size in the direction of frequency axis.

        """
        super(UpsampleNetwork, self).__init__()
        self.use_causal_conv = use_causal_conv
        self.up_layers = torch.nn.ModuleList()
        for scale in upsample_scales:
            # interpolation layer
            stretch = Stretch2d(scale, 1, interpolate_mode)
            self.up_layers += [stretch]

            # conv layer
            assert (
                freq_axis_kernel_size - 1
            ) % 2 == 0, "Not support even number freq axis kernel size."
            freq_axis_padding = (freq_axis_kernel_size - 1) // 2
            kernel_size = (freq_axis_kernel_size, scale * 2 + 1)
            if use_causal_conv:
                padding = (freq_axis_padding, scale * 2)
            else:
                padding = (freq_axis_padding, scale)
            conv = Conv2d(1, 1, kernel_size=kernel_size, padding=padding, bias=False)
            self.up_layers += [conv]

            # nonlinear
            if nonlinear_activation is not None:
                nonlinear = getattr(torch.nn, nonlinear_activation)(
                    **nonlinear_activation_params
                )
                self.up_layers += [nonlinear]

    def forward(self, c):
        """Calculate forward propagation.

        Args:
            c : Input tensor (B, C, T).

        Returns:
            Tensor: Upsampled tensor (B, C, T'), where T' = T * prod(upsample_scales).

        """
        c = c.unsqueeze(1)  # (B, 1, C, T)
        for f in self.up_layers:
            if self.use_causal_conv and isinstance(f, Conv2d):
                c = f(c)[..., : c.size(-1)]
            else:
                c = f(c)

        return c.squeeze(1)  # (B, C, T')


class ConvInUpsampleNetwork(torch.nn.Module):
    """Convolution + upsampling network module."""

    def __init__(
        self,
        upsample_scales,
        nonlinear_activation=None,
        nonlinear_activation_params={},
        interpolate_mode="nearest",
        freq_axis_kernel_size=1,
        aux_channels=80,
        aux_context_window=0,
        use_causal_conv=False,
    ):
        """Initialize convolution + upsampling network module.

        Args:
            upsample_scales (list): List of upsampling scales.
            nonlinear_activation (str): Activation function name.
            nonlinear_activation_params (dict): Arguments for specified activation function.
            mode (str): Interpolation mode.
            freq_axis_kernel_size (int): Kernel size in the direction of frequency axis.
            aux_channels (int): Number of channels of pre-convolutional layer.
            aux_context_window (int): Context window size of the pre-convolutional layer.
            use_causal_conv (bool): Whether to use causal structure.

        """
        super(ConvInUpsampleNetwork, self).__init__()
        self.aux_context_window = aux_context_window
        self.use_causal_conv = use_causal_conv and aux_context_window > 0
        # To capture wide-context information in conditional features
        kernel_size = (
            aux_context_window + 1 if use_causal_conv else 2 * aux_context_window + 1
        )
        # NOTE(kan-bayashi): Here do not use padding because the input is already padded
        self.conv_in = Conv1d(
            aux_channels, aux_channels, kernel_size=kernel_size, bias=False
        )
        self.upsample = UpsampleNetwork(
            upsample_scales=upsample_scales,
            nonlinear_activation=nonlinear_activation,
            nonlinear_activation_params=nonlinear_activation_params,
            interpolate_mode=interpolate_mode,
            freq_axis_kernel_size=freq_axis_kernel_size,
            use_causal_conv=use_causal_conv,
        )

    def forward(self, c):
        """Calculate forward propagation.

        Args:
            c : Input tensor (B, C, T').

        Returns:
            Tensor: Upsampled tensor (B, C, T),
                where T = (T' - aux_context_window * 2) * prod(upsample_scales).

        Note:
            The length of inputs considers the context window size.

        """
        c_ = self.conv_in(c)
        c = c_[:, :, : -self.aux_context_window] if self.use_causal_conv else c_
        return self.upsample(c)



import os
from os.path import exists, join

import hydra
import joblib
import numpy as np
import pyworld
import torch
from hydra.utils import to_absolute_path
from nnsvs.acoustic_models.util import pad_inference
from nnsvs.base import PredictionType
from nnsvs.gen import get_windows
from nnsvs.logger import getLogger
from nnsvs.mdn import mdn_get_most_probable_sigma_and_mu
from nnsvs.multistream import (
    get_static_features,
    get_static_stream_sizes,
    multi_stream_mlpg,
    split_streams,
)
from nnsvs.postfilters import variance_scaling
from nnsvs.util import StandardScaler, load_utt_list
from omegaconf import DictConfig, OmegaConf
from tqdm import tqdm

logger = None

use_cuda = torch.cuda.is_available()


@torch.no_grad()
def _gen_static_features(
    model, model_config, in_feats, gt_feats, out_scaler, gta=False
):
    if model.prediction_type() == PredictionType.PROBABILISTIC:
        if gta:
            raise ValueError("GTA not supported for probabilistic models for now")
        else:
            max_mu, max_sigma = model.inference(in_feats, [in_feats.shape[1]])

        if np.any(model_config.has_dynamic_features):
            # Apply denormalization
            # (B, T, D_out) -> (T, D_out)
            max_sigma_sq = (
                max_sigma.squeeze(0).cpu().data.numpy() ** 2 * out_scaler.var_
            )
            max_mu = out_scaler.inverse_transform(max_mu.squeeze(0).cpu().data.numpy())
            # Apply MLPG
            # (T, D_out) -> (T, static_dim)
            out_feats = multi_stream_mlpg(
                max_mu,
                max_sigma_sq,
                get_windows(model_config.num_windows),
                model_config.stream_sizes,
                model_config.has_dynamic_features,
            )
        else:
            # (T, D_out)
            out_feats = max_mu.squeeze(0).cpu().data.numpy()
            out_feats = out_scaler.inverse_transform(out_feats)
    else:
        if gta:
            if hasattr(model, "reduction_factor"):
                out_feats = pad_inference(
                    model,
                    in_feats,
                    [in_feats.shape[1]],
                    model.reduction_factor,
                    mode="replicate",
                    y=gt_feats,
                    mdn=model.prediction_type() in [PredictionType.PROBABILISTIC, 3],
                )
                if model.prediction_type() == PredictionType.PROBABILISTIC:
                    # (sigma, mu) -> (mu)
                    out_feats = mdn_get_most_probable_sigma_and_mu(out_feats)[1]
                elif model.prediction_type() == 3:
                    # (mu, sigma) -> (mu)
                    out_feats = out_feats[0]
            else:
                out_feats = model(in_feats, [in_feats.shape[1]], gt_feats)
                # out + residual case
                out_feats = out_feats[0] if isinstance(out_feats, tuple) else out_feats
                # multiple outputs case
                out_feats = out_feats[-1] if isinstance(out_feats, list) else out_feats
            out_feats = out_feats.squeeze(0).cpu().data.numpy()
        else:
            out_feats = (
                model.inference(in_feats, [in_feats.shape[1]])
                .squeeze(0)
                .cpu()
                .data.numpy()
            )
        out_feats = out_scaler.inverse_transform(out_feats)

        # Apply MLPG if necessary
        if np.any(model_config.has_dynamic_features):
            out_feats = multi_stream_mlpg(
                out_feats,
                out_scaler.var_,
                get_windows(model_config.num_windows),
                model_config.stream_sizes,
                model_config.has_dynamic_features,
            )

    return out_feats.astype(np.float32)


@hydra.main(config_path="conf/gen_static_features", config_name="config")
def my_app(config: DictConfig) -> None:
    global logger
    logger = getLogger(config.verbose)
    logger.info(OmegaConf.to_yaml(config))

    device = torch.device("cuda" if use_cuda else "cpu")
    utt_list = to_absolute_path(config.utt_list)
    in_dir = to_absolute_path(config.in_dir)
    gt_dir = to_absolute_path(config.gt_dir)
    out_dir = to_absolute_path(config.out_dir)

    utt_ids = load_utt_list(utt_list)

    os.makedirs(out_dir, exist_ok=True)

    model_config = OmegaConf.load(to_absolute_path(config.model.model_yaml))
    model = hydra.utils.instantiate(model_config.netG).to(device)
    checkpoint = torch.load(
        to_absolute_path(config.model.checkpoint),
        map_location=lambda storage, loc: storage,
    )
    model.load_state_dict(checkpoint["state_dict"])
    model.eval()

    out_scaler = joblib.load(to_absolute_path(config.out_scaler_path))

    mean_ = get_static_features(
        out_scaler.mean_.reshape(1, 1, out_scaler.mean_.shape[-1]),
        model_config.num_windows,
        model_config.stream_sizes,
        model_config.has_dynamic_features,
    )
    mean_ = np.concatenate(mean_, -1).reshape(1, -1)
    var_ = get_static_features(
        out_scaler.var_.reshape(1, 1, out_scaler.var_.shape[-1]),
        model_config.num_windows,
        model_config.stream_sizes,
        model_config.has_dynamic_features,
    )
    var_ = np.concatenate(var_, -1).reshape(1, -1)
    scale_ = get_static_features(
        out_scaler.scale_.reshape(1, 1, out_scaler.scale_.shape[-1]),
        model_config.num_windows,
        model_config.stream_sizes,
        model_config.has_dynamic_features,
    )
    scale_ = np.concatenate(scale_, -1).reshape(1, -1)
    static_scaler = StandardScaler(mean_, var_, scale_)

    static_stream_sizes = get_static_stream_sizes(
        model_config.stream_sizes,
        model_config.has_dynamic_features,
        model_config.num_windows,
    )

    for utt_id in tqdm(utt_ids):
        if not exists(join(gt_dir, utt_id + "-feats.npy")):
            print(f"Skip {utt_id}")
            continue
        in_feats = (
            torch.from_numpy(np.load(join(in_dir, utt_id + "-feats.npy")))
            .unsqueeze(0)
            .to(device)
        )
        gt_feats = (
            torch.from_numpy(np.load(join(gt_dir, utt_id + "-feats.npy")))
            .unsqueeze(0)
            .to(device)
        )
        static_feats = _gen_static_features(
            model, model_config, in_feats, gt_feats, out_scaler, config.gta
        )
        outs = split_streams(static_feats, static_stream_sizes)

        if len(outs) == 4:
            mgc, lf0, vuv, bap = outs

            mgc_end_dim = static_stream_sizes[0]
            bap_start_dim = sum(static_stream_sizes[:3])
            bap_end_dim = sum(static_stream_sizes[:4])

            if config.gv_postfilter:
                # mgc
                mgc = variance_scaling(
                    static_scaler.var_.reshape(-1)[:mgc_end_dim],
                    mgc,
                    offset=config.mgc_offset,
                )
                # bap
                bap = variance_scaling(
                    static_scaler.var_.reshape(-1)[bap_start_dim:bap_end_dim],
                    bap,
                    offset=config.bap_offset,
                )

            if config.mgc2sp:
                # 24k: 1024
                # 48: 2048
                fft_size = pyworld.get_cheaptrick_fft_size(config.sample_rate)
                sp = np.log(
                    pyworld.decode_spectral_envelope(
                        mgc.astype(np.float64), config.sample_rate, fft_size
                    ).astype(np.float32)
                )
                static_feats = np.concatenate([sp, lf0, vuv, bap], axis=-1)
            else:
                static_feats = np.concatenate([mgc, lf0, vuv, bap], axis=-1)
        elif len(outs) == 3:
            mel, lf0, vuv = outs
            # NOTE: may add postfilter here
            static_feats = np.concatenate([mel, lf0, vuv], axis=-1)
        else:
            raise ValueError("Invalid number of streams")

        gt_feats = gt_feats.squeeze(0)
        if len(static_feats) != len(gt_feats):
            print(static_feats.shape, gt_feats.shape)
            raise RuntimeError(f"Length mismatch in {utt_id}")

        if config.normalize:
            assert not config.mgc2sp, "need to compute normalization stats"
            static_feats = static_scaler.transform(static_feats)
        out_path = join(out_dir, f"{utt_id}-feats.npy")
        np.save(out_path, static_feats.astype(np.float32), allow_pickle=False)


def entry():
    my_app()


if __name__ == "__main__":
    my_app()



from torch import nn
from torch.nn.utils import weight_norm


def WNConv1d(*args, **kwargs):
    return weight_norm(nn.Conv1d(*args, **kwargs))


class ResnetBlock(nn.Module):
    def __init__(self, dim, dilation=1):
        super().__init__()
        self.block = nn.Sequential(
            nn.LeakyReLU(0.2),
            nn.ReflectionPad1d(dilation),
            WNConv1d(dim, dim, kernel_size=3, dilation=dilation),
            nn.LeakyReLU(0.2),
            WNConv1d(dim, dim, kernel_size=1),
        )
        self.shortcut = WNConv1d(dim, dim, kernel_size=1)

    def forward(self, x):
        return self.shortcut(x) + self.block(x)



# Acknowledgement: some of the code was adapted from ESPnet
#  Copyright 2019 Nagoya University (Tomoki Hayashi)
#  Apache 2.0  (http://www.apache.org/licenses/LICENSE-2.0)


from torch import nn


class Postnet(nn.Module):
    """Post-Net of Tacotron 2

    Args:
        in_dim (int): dimension of input
        layers (int): number of layers
        channels (int): number of channels
        kernel_size (int): kernel size
        dropout (float): dropout rate
    """

    def __init__(
        self,
        in_dim,
        layers=5,
        channels=512,
        kernel_size=5,
        dropout=0.5,
    ):
        super().__init__()
        postnet = nn.ModuleList()
        for layer in range(layers):
            in_channels = in_dim if layer == 0 else channels
            out_channels = in_dim if layer == layers - 1 else channels
            postnet += [
                nn.Conv1d(
                    in_channels,
                    out_channels,
                    kernel_size,
                    stride=1,
                    padding=(kernel_size - 1) // 2,
                    bias=False,
                ),
                nn.BatchNorm1d(out_channels),
            ]
            if layer != layers - 1:
                postnet += [nn.Tanh()]
            postnet += [nn.Dropout(dropout)]
        self.postnet = nn.Sequential(*postnet)

    def forward(self, xs):
        """Forward step

        Args:
            xs (torch.Tensor): input sequence

        Returns:
            torch.Tensor: output sequence
        """
        return self.postnet(xs)



import torch
from nnsvs.base import PredictionType
from nnsvs.util import init_seed


def _test_model_impl(model, in_dim, out_dim):
    B = 4
    T = 100
    init_seed(B * T)
    x = torch.rand(B, T, in_dim)
    y = torch.rand(B, T, out_dim)
    lengths = torch.Tensor([T] * B).long()

    # warmup forward pass
    with torch.no_grad():
        outs = model(x, lengths, y)
        if model.has_residual_lf0_prediction():
            y, lf0_residual = outs
        else:
            y, lf0_residual = outs, None
        y_inf = model.inference(x, lengths)

    # Hybrid (MDN + non-MDN)
    if model.prediction_type() == PredictionType.MULTISTREAM_HYBRID:
        # TODO
        pass
    # MDN case
    elif model.prediction_type() == PredictionType.PROBABILISTIC:
        log_pi, log_sigma, mu = y
        num_gaussian = log_pi.shape[2]
        assert mu.shape == (B, T, num_gaussian, out_dim)
        assert log_sigma.shape == (B, T, num_gaussian, out_dim)
        if lf0_residual is not None:
            assert lf0_residual.shape == (B, T, num_gaussian)
        # NOTE: infernece output shouldn't have num_gaussian axis
        mu_inf, sigma_inf = y_inf
        assert mu_inf.shape == (B, T, out_dim)
        assert sigma_inf.shape == (B, T, out_dim)
    else:
        if lf0_residual is not None:
            if isinstance(lf0_residual, list):
                lf0_residual = lf0_residual[-1]
            assert lf0_residual.shape == (B, T, 1)
        # NOTE: some models have multiple outputs (e.g. Tacotron)
        if isinstance(y, list):
            y = y[-1]
        assert y.shape == (B, T, out_dim)
        assert y.shape == y_inf.shape



import torch
from nnsvs.acoustic_models.util import pad_inference
from nnsvs.base import BaseModel
from nnsvs.tacotron.decoder import MDNNonAttentiveDecoder
from nnsvs.tacotron.decoder import NonAttentiveDecoder as TacotronNonAttentiveDecoder
from nnsvs.tacotron.postnet import Postnet as TacotronPostnet
from nnsvs.util import init_weights
from torch import nn
from torch.nn.utils.rnn import pack_padded_sequence, pad_packed_sequence

__all__ = [
    "NonAttentiveDecoder",
    "MDNNonAttentiveDecoder",
    "BiLSTMNonAttentiveDecoder",
    "BiLSTMMDNNonAttentiveDecoder",
]


class NonAttentiveDecoder(TacotronNonAttentiveDecoder):
    """Non-attentive autoregresive model based on the duration-informed Tacotron

    Duration-informed Tacotron :cite:t:`okamoto2019tacotron`.

    .. note::

        if the target features of the decoder is normalized to N(0, 1), consider
        setting the initial value carefully so that it roughly matches the value of
        silence. e.g., -4 to -10.
        ``initial_value=0`` works okay for large databases but I found that -4 or
        lower worked better for smaller databases such as nit-song070.

    Args:
        in_dim (int): Input dimension.
        out_dim (int): Output dimension.
        layers (int): Number of LSTM layers.
        hidden_dim (int): Hidden dimension of LSTM.
        prenet_layers (int): Number of prenet layers.
        prenet_hidden_dim (int): Hidden dimension of prenet.
        prenet_dropout (float): Dropout rate of prenet.
        zoneout (float): Zoneout rate.
        reduction_factor (int): Reduction factor.
        downsample_by_conv (bool): If True, downsampling is performed by convolution.
        postnet_layers (int): Number of postnet layers.
        postnet_channels (int): Number of postnet channels.
        postnet_kernel_size (int): Kernel size of postnet.
        postnet_dropout (float): Dropout rate of postnet.
        init_type (str): Initialization type.
        eval_dropout (bool): If True, dropout is applied in evaluation.
        initial_value (float) : initial value for the autoregressive decoder.
    """

    def __init__(
        self,
        in_dim=512,
        out_dim=80,
        layers=2,
        hidden_dim=1024,
        prenet_layers=2,
        prenet_hidden_dim=256,
        prenet_dropout=0.5,
        zoneout=0.1,
        reduction_factor=1,
        downsample_by_conv=False,
        postnet_layers=0,
        postnet_channels=512,
        postnet_kernel_size=5,
        postnet_dropout=0.0,
        init_type="none",
        eval_dropout=True,
        prenet_noise_std=0.0,
        initial_value=0.0,
    ):
        super().__init__(
            in_dim=in_dim,
            out_dim=out_dim,
            layers=layers,
            hidden_dim=hidden_dim,
            prenet_layers=prenet_layers,
            prenet_hidden_dim=prenet_hidden_dim,
            prenet_dropout=prenet_dropout,
            zoneout=zoneout,
            reduction_factor=reduction_factor,
            downsample_by_conv=downsample_by_conv,
            eval_dropout=eval_dropout,
            prenet_noise_std=prenet_noise_std,
            initial_value=initial_value,
        )
        if postnet_layers > 0:
            self.postnet = TacotronPostnet(
                out_dim,
                layers=postnet_layers,
                channels=postnet_channels,
                kernel_size=postnet_kernel_size,
                dropout=postnet_dropout,
            )
        else:
            self.postnet = None
        init_weights(self, init_type)

    def forward(self, x, lengths=None, y=None):
        outs = super().forward(x, lengths, y)

        if self.postnet is not None:
            # NOTE: `outs.clone()`` is necessary to compute grad on both outs and outs_fine
            outs_fine = outs + self.postnet(outs.transpose(1, 2).clone()).transpose(
                1, 2
            )
            return [outs, outs_fine]
        else:
            return outs

    def inference(self, x, lengths=None):
        return pad_inference(
            model=self, x=x, lengths=lengths, reduction_factor=self.reduction_factor
        )


class BiLSTMNonAttentiveDecoder(BaseModel):
    """BiLSTM-based encoder + NonAttentiveDecoder

    The encoder is based on the arthitecture of the Sinsy acoustic model.

    Args:
        in_dim (int): Input dimension.
        ff_hidden_dim (int): Hidden dimension of feed-forward layers in the encoder.
        conv_hidden_dim (int): Hidden dimension of convolution layers in the encoder.
        lstm_hidden_dim (int): Hidden dimension of LSTM layers in the encoder.
        num_lstm_layers (int): Number of LSTM layers in the encoder.
        out_dim (int): Output dimension.
        layers (int): Number of LSTM layers.
        hidden_dim (int): Hidden dimension of LSTM.
        prenet_layers (int): Number of prenet layers.
        prenet_hidden_dim (int): Hidden dimension of prenet.
        prenet_dropout (float): Dropout rate of prenet.
        zoneout (float): Zoneout rate.
        reduction_factor (int): Reduction factor.
        downsample_by_conv (bool): If True, downsampling is performed by convolution.
        postnet_layers (int): Number of postnet layers.
        postnet_channels (int): Number of postnet channels.
        postnet_kernel_size (int): Kernel size of postnet.
        postnet_dropout (float): Dropout rate of postnet.
        in_ph_start_idx (int): Start index of phoneme features.
        in_ph_end_idx (int): End index of phoneme features.
        embed_dim (int): Embedding dimension.
        init_type (str): Initialization type.
        eval_dropout (bool): If True, dropout is applied in evaluation.
        initial_value (float) : initial value for the autoregressive decoder.
    """

    def __init__(
        self,
        in_dim=512,
        ff_hidden_dim=2048,
        conv_hidden_dim=1024,
        lstm_hidden_dim=256,
        num_lstm_layers=2,
        out_dim=80,
        decoder_layers=2,
        decoder_hidden_dim=1024,
        prenet_layers=2,
        prenet_hidden_dim=256,
        prenet_dropout=0.5,
        zoneout=0.1,
        reduction_factor=1,
        downsample_by_conv=False,
        postnet_layers=0,
        postnet_channels=512,
        postnet_kernel_size=5,
        postnet_dropout=0.0,
        in_ph_start_idx: int = 1,
        in_ph_end_idx: int = 50,
        embed_dim=None,
        init_type="none",
        eval_dropout=True,
        prenet_noise_std=0.0,
        initial_value=0.0,
    ):
        super().__init__()
        self.in_dim = in_dim
        self.out_dim = out_dim
        self.in_ph_start_idx = in_ph_start_idx
        self.in_ph_end_idx = in_ph_end_idx
        self.num_vocab = in_ph_end_idx - in_ph_start_idx
        self.embed_dim = embed_dim
        self.reduction_factor = reduction_factor

        if self.embed_dim is not None:
            assert in_dim > self.num_vocab
            self.emb = nn.Embedding(self.num_vocab, embed_dim)
            self.fc_in = nn.Linear(in_dim - self.num_vocab, embed_dim)
            ff_in_dim = embed_dim
        else:
            ff_in_dim = in_dim

        # Encoder
        # NOTE: can be simply replaced by a BiLSTM?
        # so far I use sinsy like architecture
        self.ff = nn.Sequential(
            nn.Linear(ff_in_dim, ff_hidden_dim),
            nn.ReLU(),
            nn.Linear(ff_hidden_dim, ff_hidden_dim),
            nn.ReLU(),
            nn.Linear(ff_hidden_dim, ff_hidden_dim),
            nn.ReLU(),
        )

        self.conv = nn.Sequential(
            nn.ReflectionPad1d(3),
            nn.Conv1d(ff_hidden_dim, conv_hidden_dim, kernel_size=7, padding=0),
            nn.BatchNorm1d(conv_hidden_dim),
            nn.ReLU(),
            nn.ReflectionPad1d(3),
            nn.Conv1d(conv_hidden_dim, conv_hidden_dim, kernel_size=7, padding=0),
            nn.BatchNorm1d(conv_hidden_dim),
            nn.ReLU(),
            nn.ReflectionPad1d(3),
            nn.Conv1d(conv_hidden_dim, conv_hidden_dim, kernel_size=7, padding=0),
            nn.BatchNorm1d(conv_hidden_dim),
            nn.ReLU(),
        )

        self.lstm = nn.LSTM(
            conv_hidden_dim,
            lstm_hidden_dim,
            num_lstm_layers,
            bidirectional=True,
            batch_first=True,
            dropout=0.0,
        )

        # Autoregressive decoder
        decoder_in_dim = 2 * lstm_hidden_dim
        self.decoder = TacotronNonAttentiveDecoder(
            in_dim=decoder_in_dim,
            out_dim=out_dim,
            layers=decoder_layers,
            hidden_dim=decoder_hidden_dim,
            prenet_layers=prenet_layers,
            prenet_hidden_dim=prenet_hidden_dim,
            prenet_dropout=prenet_dropout,
            zoneout=zoneout,
            reduction_factor=reduction_factor,
            downsample_by_conv=downsample_by_conv,
            eval_dropout=eval_dropout,
            prenet_noise_std=prenet_noise_std,
            initial_value=initial_value,
        )

        if postnet_layers > 0:
            self.postnet = TacotronPostnet(
                out_dim,
                layers=postnet_layers,
                channels=postnet_channels,
                kernel_size=postnet_kernel_size,
                dropout=postnet_dropout,
            )
        else:
            self.postnet = None
        init_weights(self, init_type)

    def is_autoregressive(self):
        return self.decoder.is_autoregressive()

    def forward(self, x, lengths=None, y=None):
        if isinstance(lengths, torch.Tensor):
            lengths = lengths.to("cpu")

        if self.embed_dim is not None:
            x_first, x_ph_onehot, x_last = torch.split(
                x,
                [
                    self.in_ph_start_idx,
                    self.num_vocab,
                    self.in_dim - self.num_vocab - self.in_ph_start_idx,
                ],
                dim=-1,
            )
            x_ph = torch.argmax(x_ph_onehot, dim=-1)
            # Make sure to have one-hot vector
            assert (x_ph_onehot.sum(-1) <= 1).all()
            x = self.emb(x_ph) + self.fc_in(torch.cat([x_first, x_last], dim=-1))

        out = self.ff(x)
        out = self.conv(out.transpose(1, 2)).transpose(1, 2)
        sequence = pack_padded_sequence(out, lengths, batch_first=True)
        out, _ = self.lstm(sequence)
        out, _ = pad_packed_sequence(out, batch_first=True)
        outs = self.decoder(out, lengths, y)

        if self.postnet is not None:
            # NOTE: `outs.clone()`` is necessary to compute grad on both outs and outs_fine
            outs_fine = outs + self.postnet(outs.transpose(1, 2).clone()).transpose(
                1, 2
            )
            return [outs, outs_fine]
        else:
            return outs

    def inference(self, x, lengths=None):
        return pad_inference(
            model=self, x=x, lengths=lengths, reduction_factor=self.reduction_factor
        )


class BiLSTMMDNNonAttentiveDecoder(BaseModel):
    """BiLSTM-based encoder + NonAttentiveDecoder (MDN version)


    The encoder is based on the arthitecture of the Sinsy acoustic model.

    Args:
        in_dim (int): Input dimension.
        ff_hidden_dim (int): Hidden dimension of feed-forward layers in the encoder.
        conv_hidden_dim (int): Hidden dimension of convolution layers in the encoder.
        lstm_hidden_dim (int): Hidden dimension of LSTM layers in the encoder.
        num_lstm_layers (int): Number of LSTM layers in the encoder.
        out_dim (int): Output dimension.
        layers (int): Number of LSTM layers.
        hidden_dim (int): Hidden dimension of LSTM.
        prenet_layers (int): Number of prenet layers.
        prenet_hidden_dim (int): Hidden dimension of prenet.
        prenet_dropout (float): Dropout rate of prenet.
        zoneout (float): Zoneout rate.
        reduction_factor (int): Reduction factor.
        downsample_by_conv (bool): If True, downsampling is performed by convolution.
        num_gaussians (int): Number of Gaussians.
        sampling_mode (str): Sampling mode.
        postnet_layers (int): Number of postnet layers.
        postnet_channels (int): Number of postnet channels.
        postnet_kernel_size (int): Kernel size of postnet.
        postnet_dropout (float): Dropout rate of postnet.
        in_ph_start_idx (int): Start index of phoneme features.
        in_ph_end_idx (int): End index of phoneme features.
        embed_dim (int): Embedding dimension.
        init_type (str): Initialization type.
        eval_dropout (bool): If True, dropout is applied in evaluation.
        initial_value (float) : initial value for the autoregressive decoder.
    """

    def __init__(
        self,
        in_dim=512,
        ff_hidden_dim=2048,
        conv_hidden_dim=1024,
        lstm_hidden_dim=256,
        num_lstm_layers=2,
        out_dim=80,
        decoder_layers=2,
        decoder_hidden_dim=1024,
        prenet_layers=2,
        prenet_hidden_dim=256,
        prenet_dropout=0.5,
        zoneout=0.1,
        reduction_factor=1,
        downsample_by_conv=False,
        num_gaussians=8,
        sampling_mode="mean",
        in_ph_start_idx: int = 1,
        in_ph_end_idx: int = 50,
        embed_dim=None,
        init_type="none",
        eval_dropout=True,
        prenet_noise_std=0,
        initial_value=0.0,
    ):
        super().__init__()
        self.in_dim = in_dim
        self.out_dim = out_dim
        self.in_ph_start_idx = in_ph_start_idx
        self.in_ph_end_idx = in_ph_end_idx
        self.num_vocab = in_ph_end_idx - in_ph_start_idx
        self.embed_dim = embed_dim
        self.reduction_factor = reduction_factor

        if self.embed_dim is not None:
            assert in_dim > self.num_vocab
            self.emb = nn.Embedding(self.num_vocab, embed_dim)
            self.fc_in = nn.Linear(in_dim - self.num_vocab, embed_dim)
            ff_in_dim = embed_dim
        else:
            ff_in_dim = in_dim

        # Encoder
        # NOTE: can be simply replaced by a BiLSTM?
        # so far I use sinsy like architecture
        self.ff = nn.Sequential(
            nn.Linear(ff_in_dim, ff_hidden_dim),
            nn.ReLU(),
            nn.Linear(ff_hidden_dim, ff_hidden_dim),
            nn.ReLU(),
            nn.Linear(ff_hidden_dim, ff_hidden_dim),
            nn.ReLU(),
        )

        self.conv = nn.Sequential(
            nn.ReflectionPad1d(3),
            nn.Conv1d(ff_hidden_dim, conv_hidden_dim, kernel_size=7, padding=0),
            nn.BatchNorm1d(conv_hidden_dim),
            nn.ReLU(),
            nn.ReflectionPad1d(3),
            nn.Conv1d(conv_hidden_dim, conv_hidden_dim, kernel_size=7, padding=0),
            nn.BatchNorm1d(conv_hidden_dim),
            nn.ReLU(),
            nn.ReflectionPad1d(3),
            nn.Conv1d(conv_hidden_dim, conv_hidden_dim, kernel_size=7, padding=0),
            nn.BatchNorm1d(conv_hidden_dim),
            nn.ReLU(),
        )

        self.lstm = nn.LSTM(
            conv_hidden_dim,
            lstm_hidden_dim,
            num_lstm_layers,
            bidirectional=True,
            batch_first=True,
            dropout=0.0,
        )

        # Autoregressive decoder
        decoder_in_dim = 2 * lstm_hidden_dim
        self.decoder = MDNNonAttentiveDecoder(
            in_dim=decoder_in_dim,
            out_dim=out_dim,
            layers=decoder_layers,
            hidden_dim=decoder_hidden_dim,
            prenet_layers=prenet_layers,
            prenet_hidden_dim=prenet_hidden_dim,
            prenet_dropout=prenet_dropout,
            zoneout=zoneout,
            reduction_factor=reduction_factor,
            downsample_by_conv=downsample_by_conv,
            num_gaussians=num_gaussians,
            sampling_mode=sampling_mode,
            eval_dropout=eval_dropout,
            prenet_noise_std=prenet_noise_std,
            initial_value=initial_value,
        )
        init_weights(self, init_type)

    def is_autoregressive(self):
        return self.decoder.is_autoregressive()

    def prediction_type(self):
        return self.decoder.prediction_type()

    def forward(self, x, lengths=None, y=None):
        if isinstance(lengths, torch.Tensor):
            lengths = lengths.to("cpu")

        if self.embed_dim is not None:
            x_first, x_ph_onehot, x_last = torch.split(
                x,
                [
                    self.in_ph_start_idx,
                    self.num_vocab,
                    self.in_dim - self.num_vocab - self.in_ph_start_idx,
                ],
                dim=-1,
            )
            x_ph = torch.argmax(x_ph_onehot, dim=-1)
            # Make sure to have one-hot vector
            assert (x_ph_onehot.sum(-1) <= 1).all()
            x = self.emb(x_ph) + self.fc_in(torch.cat([x_first, x_last], dim=-1))

        out = self.ff(x)
        out = self.conv(out.transpose(1, 2)).transpose(1, 2)
        sequence = pack_padded_sequence(out, lengths, batch_first=True)
        out, _ = self.lstm(sequence)
        out, _ = pad_packed_sequence(out, batch_first=True)
        outs = self.decoder(out, lengths, y)

        return outs

    def inference(self, x, lengths=None):
        return pad_inference(
            model=self,
            x=x,
            lengths=lengths,
            reduction_factor=self.reduction_factor,
            mdn=True,
        )



from .wavenet import WaveNet

__all__ = ["WaveNet", "receptive_field_size"]


def receptive_field_size(
    total_layers, num_cycles, kernel_size, dilation=lambda x: 2 ** x
):
    """Compute receptive field size of WaveNet

    Args:
        total_layers (int): total layers
        num_cycles (int): cycles
        kernel_size (int): kernel size
        dilation (lambda): lambda to compute dilation factor. ``lambda x : 1``
          to disable dilated convolution.

    Returns:
        int: receptive field size in sample

    Examples:

    .. ipython::

        In [1]: from ttslearn.wavenet import receptive_field_size

        In [2]: receptive_field_size(30, 3, 2)
        Out[2]: 3070
    """
    assert total_layers % num_cycles == 0
    layers_per_cycle = total_layers // num_cycles
    dilations = [dilation(i % layers_per_cycle) for i in range(total_layers)]
    return (kernel_size - 1) * sum(dilations) + 1



def test_ja():
    from nnsvs.frontend import ja

    print("num vocab:", ja.num_vocab())
    seq = ja.text_to_sequence(["a", "i", "u", "e", "o"])
    assert " ".join(ja.sequence_to_text(seq)) == "a i u e o"


def test_zh():
    from nnsvs.frontend import zh

    print("num vocab:", zh.num_vocab())
    seq = zh.text_to_sequence(["AP", "SP"])
    assert " ".join(zh.sequence_to_text(seq)) == "AP SP"



# -*- coding: utf-8 -*-

# Copyright 2022 Reo Yoneyama (Nagoya University)
#  MIT License (https://opensource.org/licenses/MIT)

"""Feature-related functions.

References:
    - https://github.com/bigpon/QPPWG

"""

import sys
from logging import getLogger

import numpy as np
import torch
from torch.nn.functional import interpolate

# A logger for this file
logger = getLogger(__name__)


def validate_length(x, y, hop_size=None):
    """Validate length

    Args:
        x (ndarray): numpy array with x.shape[0] = len_x
        y (ndarray): numpy array with y.shape[0] = len_y
        hop_size (int): upsampling factor

    Returns:
        (ndarray): length adjusted x with same length y
        (ndarray): length adjusted y with same length x

    """
    if hop_size is None:
        if x.shape[0] < y.shape[0]:
            y = y[: x.shape[0]]
        if x.shape[0] > y.shape[0]:
            x = x[: y.shape[0]]
        assert len(x) == len(y)
    else:
        if x.shape[0] > y.shape[0] * hop_size:
            x = x[: y.shape[0] * hop_size]
        if x.shape[0] < y.shape[0] * hop_size:
            mod_y = y.shape[0] * hop_size - x.shape[0]
            mod_y_frame = mod_y // hop_size + 1
            y = y[:-mod_y_frame]
            x = x[: y.shape[0] * hop_size]
        assert len(x) == len(y) * hop_size

    return x, y


def dilated_factor(batch_f0, fs, dense_factor):
    """Pitch-dependent dilated factor

    Args:
        batch_f0 (ndarray): the f0 sequence (T)
        fs (int): sampling rate
        dense_factor (int): the number of taps in one cycle

    Return:
        dilated_factors(np array):
            float array of the pitch-dependent dilated factors (T)

    """
    batch_f0[batch_f0 == 0] = fs / dense_factor
    dilated_factors = np.ones(batch_f0.shape) * fs
    dilated_factors /= batch_f0
    dilated_factors /= dense_factor
    assert np.all(dilated_factors > 0)

    return dilated_factors


class SignalGenerator:
    """Input signal generator module."""

    def __init__(
        self,
        sample_rate=24000,
        hop_size=120,
        sine_amp=0.1,
        noise_amp=0.003,
        signal_types=["sine", "noise"],
    ):
        """Initialize WaveNetResidualBlock module.

        Args:
            sample_rate (int): Sampling rate.
            hop_size (int): Hop size of input F0.
            sine_amp (float): Sine amplitude for NSF-based sine generation.
            noise_amp (float): Noise amplitude for NSF-based sine generation.
            signal_types (list): List of input signal types for generator.

        """
        self.sample_rate = sample_rate
        self.hop_size = hop_size
        self.signal_types = signal_types
        self.sine_amp = sine_amp
        self.noise_amp = noise_amp

        # for signal_type in signal_types:
        #     if not signal_type in ["noise", "sine", "uv"]:
        #         logger.info(f"{signal_type} is not supported type for generator input.")
        #         sys.exit(0)
        # logger.info(f"Use {signal_types} for generator input signals.")

    @torch.no_grad()
    def __call__(self, f0):
        signals = []
        for typ in self.signal_types:
            if "noise" == typ:
                signals.append(self.random_noise(f0))
            if "sine" == typ:
                signals.append(self.sinusoid(f0))
            if "uv" == typ:
                signals.append(self.vuv_binary(f0))

        input_batch = signals[0]
        for signal in signals[1:]:
            input_batch = torch.cat([input_batch, signal], axis=1)

        return input_batch

    @torch.no_grad()
    def random_noise(self, f0):
        """Calculate noise signals.

        Args:
            f0 (Tensor): F0 tensor (B, 1, T // hop_size).

        Returns:
            Tensor: Gaussian noise signals (B, 1, T).

        """
        B, _, T = f0.size()
        noise = torch.randn((B, 1, T * self.hop_size), device=f0.device)

        return noise

    @torch.no_grad()
    def sinusoid(self, f0):
        """Calculate sine signals.

        Args:
            f0 (Tensor): F0 tensor (B, 1, T // hop_size).

        Returns:
            Tensor: Sines generated following NSF (B, 1, T).

        """
        B, _, T = f0.size()
        vuv = interpolate((f0 > 0) * torch.ones_like(f0), T * self.hop_size)
        radious = (interpolate(f0, T * self.hop_size) / self.sample_rate) % 1
        sine = vuv * torch.sin(torch.cumsum(radious, dim=2) * 2 * np.pi) * self.sine_amp
        if self.noise_amp > 0:
            noise_amp = vuv * self.noise_amp + (1.0 - vuv) * self.noise_amp / 3.0
            noise = torch.randn((B, 1, T * self.hop_size), device=f0.device) * noise_amp
            sine = sine + noise

        return sine

    @torch.no_grad()
    def vuv_binary(self, f0):
        """Calculate V/UV binary sequences.

        Args:
            f0 (Tensor): F0 tensor (B, 1, T // hop_size).

        Returns:
            Tensor: V/UV binary sequences (B, 1, T).

        """
        _, _, T = f0.size()
        uv = interpolate((f0 > 0) * torch.ones_like(f0), T * self.hop_size)

        return uv



import numpy as np
import torch
from nnsvs.usfgan.utils import SignalGenerator, dilated_factor
from torch import nn


class USFGANWrapper(nn.Module):
    def __init__(self, config, generator):
        super().__init__()
        self.generator = generator
        self.config = config

    def inference(self, f0, aux_feats):
        """Inference for USFGAN

        Args:
            f0 (numpy.ndarray): F0 (T, 1)
            aux_feats (Tensor): Auxiliary features (T, C)

        """
        signal_generator = SignalGenerator(
            sample_rate=self.config.data.sample_rate,
            hop_size=self.config.data.hop_size,
            sine_amp=self.config.data.sine_amp,
            noise_amp=self.config.data.noise_amp,
            signal_types=self.config.data.signal_types,
        )
        assert self.config.data.sine_f0_type in ["contf0", "cf0", "f0"]
        assert self.config.data.df_f0_type in ["contf0", "cf0", "f0"]

        device = aux_feats.device

        is_sifigan = "aux_context_window" not in self.config.generator
        if is_sifigan:
            dfs = []
            for df, us in zip(
                self.config.data.dense_factors,
                np.cumprod(self.config.generator.upsample_scales),
            ):
                dfs += [
                    np.repeat(
                        dilated_factor(f0.copy(), self.config.data.sample_rate, df), us
                    )
                ]
            df = [
                torch.FloatTensor(np.array(df)).view(1, 1, -1).to(device) for df in dfs
            ]
            c = aux_feats.unsqueeze(0).transpose(2, 1).to(device)
        else:
            df = dilated_factor(
                np.squeeze(f0.copy()),
                self.config.data.sample_rate,
                self.config.data.dense_factor,
            )
            df = df.repeat(self.config.data.hop_size, axis=0)
            pad_fn = nn.ReplicationPad1d(self.config.generator.aux_context_window)
            c = pad_fn(aux_feats.unsqueeze(0).transpose(2, 1)).to(device)
            df = torch.FloatTensor(df).view(1, 1, -1).to(device)

        f0 = torch.FloatTensor(f0).unsqueeze(0).transpose(2, 1).to(device)

        in_signal = signal_generator(f0)
        y = self.generator(in_signal, c, df)[0]

        return y



from os.path import dirname, join

import numpy as np
import pytest
import pyworld
from nnsvs.pitch import (
    extract_smoothed_f0,
    extract_vibrato_likelihood,
    extract_vibrato_parameters,
    gen_sine_vibrato,
    hz_to_cent_based_c4,
)
from scipy.io import wavfile

__test_wav_file = join(dirname(__file__), "data", "nitech_jp_song070_f001_004.wav")


@pytest.mark.parametrize("frame_period", [5, 10])
def test_sine_vibrato(frame_period):
    sr, x = wavfile.read(__test_wav_file)

    if frame_period == 10:
        win_length = 32
        n_fft = 128
        threshold = 0.12
    elif frame_period == 5:
        win_length = 64
        n_fft = 256
        threshold = 0.12

    frame_shift = int(frame_period * 0.001 * sr)
    sr_f0 = int(sr / frame_shift)

    f0, timeaxis = pyworld.dio(x.astype(np.float64), sr, frame_period=frame_period)
    f0 = pyworld.stonemask(x.astype(np.float64), f0, timeaxis, sr)

    f0_smooth = extract_smoothed_f0(f0, sr_f0, cutoff=8)
    f0_smooth_cent = hz_to_cent_based_c4(f0_smooth)
    vibrato_likelihood = extract_vibrato_likelihood(
        f0_smooth_cent, sr_f0, win_length=win_length, n_fft=n_fft
    )
    vib_flags, m_a, m_f = extract_vibrato_parameters(
        f0_smooth_cent, vibrato_likelihood, sr_f0, threshold=threshold
    )
    # at least one vibrato section should be detected
    assert vib_flags.sum() > 0

    assert m_a.shape == (len(f0),)
    assert m_f.shape == (len(f0),)
    assert vib_flags.shape == (len(f0),)

    # Generate F0 with the vibrato parameters (m_a, m_f)
    f0_no_vib = extract_smoothed_f0(f0, sr_f0, cutoff=3)
    f0_with_vib = gen_sine_vibrato(f0_no_vib, sr_f0, m_a, m_f)
    assert f0_no_vib.shape == f0_with_vib.shape


def test_diff_vibrato():
    sr, x = wavfile.read(__test_wav_file)

    frame_period = 5
    frame_shift = int(frame_period * 0.001 * sr)
    sr_f0 = int(sr / frame_shift)

    f0, timeaxis = pyworld.dio(x.astype(np.float64), sr, frame_period=frame_period)
    f0 = pyworld.stonemask(x.astype(np.float64), f0, timeaxis, sr)

    f0_smooth = extract_smoothed_f0(f0, sr_f0, cutoff=8)
    vib = f0 - f0_smooth

    assert vib.shape == (len(f0),)



from importlib.machinery import SourceFileLoader
from os.path import exists

from setuptools import find_packages, setup

version = SourceFileLoader("nnsvs.version", "nnsvs/version.py").load_module().version

packages = find_packages()
if exists("README.md"):
    with open("README.md", "r", encoding="UTF-8") as fh:
        LONG_DESC = LONG_DESC = fh.read()
else:
    LONG_DESC = ""

setup(
    name="nnsvs",
    version=version,
    description="DNN-based singing voice synthesis library",
    long_description=LONG_DESC,
    long_description_content_type="text/markdown",
    package_data={"": ["_example_data/*"]},
    packages=packages,
    include_package_data=True,
    install_requires=[
        "numpy",
        "scipy",
        "cython",
        "torch >= 1.6.0",
        "torchaudio",
        "hydra-core >= 1.1.0",
        "hydra_colorlog >= 1.1.0",
        "librosa >= 0.7.0",
        "pysptk",
        "pyworld",
        "tensorboard",
        "nnmnkwii",
        "pysinsy",
        "pyloudnorm",
    ],
    extras_require={
        "dev": [
            # NOTE: tentative fix for https://github.com/nnsvs/nnsvs/issues/191
            "matplotlib<3.6.0",
            "seaborn",
            "mlflow",
            "optuna",
            "hydra-optuna-sweeper",
            "protobuf <= 3.20.1",
            "praat-parselmouth",
        ],
        "docs": [
            "sphinx",
            "sphinx-autobuild",
            "sphinx_rtd_theme",
            "nbsphinx>=0.8.6",
            "sphinxcontrib-bibtex",
            "sphinxcontrib-youtube",
            "Jinja2>=3.0.1,<=3.0.3",
            "pandoc",
            "ipython",
            "jupyter",
            "matplotlib>=1.5",
        ],
        "lint": [
            "pysen",
            "types-setuptools",
            "mypy<=0.910",
            "black>=19.19b0,<=20.8",
            "flake8>=3.7,<4",
            "flake8-bugbear",
            "isort>=4.3,<5.2.0",
            "click<8.1.0",
            "importlib-metadata<5.0",
        ],
        "test": ["pytest"],
    },
    entry_points={
        "console_scripts": [
            "nnsvs-prepare-features = nnsvs.bin.prepare_features:entry",
            "nnsvs-prepare-static-features = nnsvs.bin.prepare_static_features:entry",
            "nnsvs-prepare-voc-features = nnsvs.bin.prepare_voc_features:entry",
            "nnsvs-fit-scaler = nnsvs.bin.fit_scaler:entry",
            "nnsvs-preprocess-normalize = nnsvs.bin.preprocess_normalize:entry",
            "nnsvs-train = nnsvs.bin.train:entry",
            "nnsvs-train-acoustic = nnsvs.bin.train_acoustic:entry",
            "nnsvs-train-postfilter = nnsvs.bin.train_postfilter:entry",
            "nnsvs-generate = nnsvs.bin.generate:entry",
            "nnsvs-gen-static-features = nnsvs.bin.gen_static_features:entry",
            "nnsvs-synthesis = nnsvs.bin.synthesis:entry",
        ],
    },
)



import pytest
import torch
from nnsvs.base import PredictionType
from nnsvs.diffsinger.denoiser import DiffNet
from nnsvs.diffsinger.diffusion import GaussianDiffusion
from nnsvs.diffsinger.fs2 import FFTBlocks, FFTBlocksEncoder
from nnsvs.model import LSTMEncoder
from nnsvs.util import init_seed

from .util import _test_model_impl


@pytest.mark.parametrize("use_pos_embed", [False, True])
def test_fftblocks(use_pos_embed):
    model = FFTBlocks(16, 2, use_pos_embed=use_pos_embed)
    B = 4
    T = 100
    init_seed(B * T)
    x = torch.rand(B, T, 16)
    lengths = torch.Tensor([T] * B).long()
    y_hat = model(x, lengths)
    assert x.shape[1] == y_hat.shape[1]


@pytest.mark.parametrize("reduction_factor", [1, 4])
@pytest.mark.parametrize("downsample_by_conv", [False, True])
def test_fs2(reduction_factor, downsample_by_conv):
    params = {
        "in_dim": 86,
        "hidden_dim": 16,
        "out_dim": 16,
        "num_layers": 2,
        "ffn_kernel_size": 3,
        "reduction_factor": reduction_factor,
        "downsample_by_conv": downsample_by_conv,
    }
    model = FFTBlocksEncoder(**params)
    assert model.prediction_type() == PredictionType.DETERMINISTIC
    _test_model_impl(model, params["in_dim"], params["out_dim"])


def test_denoiser():
    model = DiffNet(
        in_dim=80,
        encoder_hidden_dim=12,
        residual_layers=2,
        residual_channels=4,
        dilation_cycle_length=4,
    )
    x = torch.rand(2, 1, 80, 100)
    cond = torch.rand(2, 12, 100)
    step = torch.randint(0, 100, (2,))
    y = model(x, step, cond)
    assert x.shape == y.shape


@pytest.mark.parametrize("pndm_speedup", [None])
def test_gaussian_diffusion(pndm_speedup):
    encoder = LSTMEncoder(
        in_dim=60,
        hidden_dim=2,
        out_dim=16,
        num_layers=2,
        dropout=0.5,
        init_type="none",
    )
    params = {
        "in_dim": 60,
        "out_dim": 80,
        "denoise_fn": DiffNet(
            in_dim=80,
            encoder_hidden_dim=16,
            residual_layers=2,
            residual_channels=4,
            dilation_cycle_length=4,
        ),
        "K_step": 100,
        "betas": None,
        "pndm_speedup": pndm_speedup,
    }
    model = GaussianDiffusion(**params)
    B = 4
    T = 100
    init_seed(B * T)
    x = torch.rand(B, T, model.in_dim)
    y = torch.rand(B, T, model.out_dim)
    lengths = torch.Tensor([T] * B).long()
    encoder_outs = encoder(x, lengths)
    noise, x_recon = model(encoder_outs, lengths, y)
    assert noise.shape == y.shape
    assert x_recon.shape == y.shape

    y_hat = model.inference(encoder_outs, lengths)
    assert y_hat.shape == y.shape


@pytest.mark.parametrize("pndm_speedup", [None])
def test_gaussian_diffusion_with_encoder(pndm_speedup):
    params = {
        "in_dim": 60,
        "out_dim": 80,
        "denoise_fn": DiffNet(
            in_dim=80,
            encoder_hidden_dim=16,
            residual_layers=2,
            residual_channels=4,
            dilation_cycle_length=4,
        ),
        "encoder": LSTMEncoder(
            in_dim=60,
            hidden_dim=2,
            out_dim=16,
            num_layers=2,
            dropout=0.5,
            init_type="none",
        ),
        "K_step": 100,
        "betas": None,
        "pndm_speedup": pndm_speedup,
    }
    model = GaussianDiffusion(**params)
    B = 4
    T = 100
    init_seed(B * T)
    x = torch.rand(B, T, model.encoder.in_dim)
    y = torch.rand(B, T, model.out_dim)
    lengths = torch.Tensor([T] * B).long()
    noise, x_recon = model(x, lengths, y)
    assert noise.shape == y.shape
    assert x_recon.shape == y.shape

    y_hat = model.inference(x, lengths)
    assert y_hat.shape == y.shape



from warnings import warn

import torch
from nnsvs.base import BaseModel, PredictionType
from nnsvs.dsp import TrTimeInvFIRFilter
from nnsvs.layers.conv import ResnetBlock, WNConv1d
from nnsvs.layers.layer_norm import LayerNorm
from nnsvs.mdn import MDNLayer, mdn_get_most_probable_sigma_and_mu
from nnsvs.multistream import split_streams
from nnsvs.transformer.attentions import sequence_mask
from nnsvs.transformer.encoder import Encoder as _TransformerEncoder
from nnsvs.util import init_weights
from torch import nn
from torch.nn.utils.rnn import pack_padded_sequence, pad_packed_sequence

__all__ = [
    "ExtractFromInput",
    "FFN",
    "LSTMRNN",
    "LSTMRNNSAR",
    "MDN",
    "MDNv2",
    "RMDN",
    "Conv1dResnet",
    "Conv1dResnetMDN",
    "Conv1dResnetSAR",
    "FFConvLSTM",
    "LSTMEncoder",
    "VariancePredictor",
    "TransformerEncoder",
]


class ExtractFromInput(BaseModel):
    """Extract a part of input

    This class doesn't have any learnable parameters.

    Args:
        start_idx (int): the start index of the input
        end_idx (int): the end index of the input

    """

    def __init__(self, start_idx, end_idx):
        super().__init__()
        self.start_idx = start_idx
        self.end_idx = end_idx

    def forward(self, x, lengths=None, y=None):
        """Forward pass"""
        # x: (B, T, C)
        return x[:, :, self.start_idx : self.end_idx]


class Conv1dResnet(BaseModel):
    """Conv1d + Resnet

    The model is inspired by the MelGAN's model architecture (:cite:t:`kumar2019melgan`).
    MDN layer is added if use_mdn is True.

    Args:
        in_dim (int): the dimension of the input
        hidden_dim (int): the dimension of the hidden state
        out_dim (int): the dimension of the output
        num_layers (int): the number of layers
        init_type (str): the type of weight initialization
        use_mdn (bool): whether to use MDN or not
        num_gaussians (int): the number of gaussians in MDN
        dim_wise (bool): whether to use dim-wise or not
        in_ph_start_idx (int): the start index of phoneme identity in a hed file
        in_ph_end_idx (int): the end index of phoneme identity in a hed file
        embed_dim (int): the dimension of the phoneme embedding
    """

    def __init__(
        self,
        in_dim,
        hidden_dim,
        out_dim,
        num_layers=4,
        init_type="none",
        use_mdn=False,
        num_gaussians=8,
        dim_wise=False,
        in_ph_start_idx: int = 1,
        in_ph_end_idx: int = 50,
        embed_dim=None,
        **kwargs,
    ):
        super().__init__()
        self.in_dim = in_dim
        self.out_dim = out_dim
        self.use_mdn = use_mdn
        self.in_ph_start_idx = in_ph_start_idx
        self.in_ph_end_idx = in_ph_end_idx
        self.num_vocab = in_ph_end_idx - in_ph_start_idx
        self.embed_dim = embed_dim

        if "dropout" in kwargs:
            warn(
                "dropout argument in Conv1dResnet is deprecated"
                " and will be removed in future versions"
            )

        if self.embed_dim is not None:
            assert in_dim > self.num_vocab
            self.emb = nn.Embedding(self.num_vocab, embed_dim)
            self.fc_in = nn.Linear(in_dim - self.num_vocab, embed_dim)
            conv_in_dim = embed_dim
        else:
            conv_in_dim = in_dim

        model = [
            nn.ReflectionPad1d(3),
            WNConv1d(conv_in_dim, hidden_dim, kernel_size=7, padding=0),
        ]
        for n in range(num_layers):
            model.append(ResnetBlock(hidden_dim, dilation=2 ** n))

        last_conv_out_dim = hidden_dim if use_mdn else out_dim
        model += [
            nn.LeakyReLU(0.2),
            nn.ReflectionPad1d(3),
            WNConv1d(hidden_dim, last_conv_out_dim, kernel_size=7, padding=0),
        ]

        self.model = nn.Sequential(*model)

        if self.use_mdn:
            self.mdn_layer = MDNLayer(
                in_dim=hidden_dim,
                out_dim=out_dim,
                num_gaussians=num_gaussians,
                dim_wise=dim_wise,
            )
        else:
            self.mdn_layer = None

        init_weights(self, init_type)

    def prediction_type(self):
        return (
            PredictionType.PROBABILISTIC
            if self.use_mdn
            else PredictionType.DETERMINISTIC
        )

    def forward(self, x, lengths=None, y=None):
        """Forward step

        Args:
            x (torch.Tensor): the input tensor
            lengths (torch.Tensor): the lengths of the input tensor
            y (torch.Tensor): the optional target tensor

        Returns:
            torch.Tensor: the output tensor
        """

        if self.embed_dim is not None:
            x_first, x_ph_onehot, x_last = torch.split(
                x,
                [
                    self.in_ph_start_idx,
                    self.num_vocab,
                    self.in_dim - self.num_vocab - self.in_ph_start_idx,
                ],
                dim=-1,
            )
            x_ph = torch.argmax(x_ph_onehot, dim=-1)
            # Make sure to have one-hot vector
            assert (x_ph_onehot.sum(-1) <= 1).all()
            x = self.emb(x_ph) + self.fc_in(torch.cat([x_first, x_last], dim=-1))

        out = self.model(x.transpose(1, 2)).transpose(1, 2)

        if self.use_mdn:
            return self.mdn_layer(out)
        else:
            return out

    def inference(self, x, lengths=None):
        """Inference step

        Find the most likely mean and variance if use_mdn is True

        Args:
            x (torch.Tensor): the input tensor
            lengths (torch.Tensor): the lengths of the input tensor

        Returns:
            tuple: mean and variance of the output features
        """
        if self.use_mdn:
            log_pi, log_sigma, mu = self(x, lengths)
            sigma, mu = mdn_get_most_probable_sigma_and_mu(log_pi, log_sigma, mu)
            return mu, sigma
        else:
            return self(x, lengths)


@torch.no_grad()
def _shallow_ar_inference(out, stream_sizes, analysis_filts):
    from torchaudio.functional import lfilter

    out_streams = split_streams(out, stream_sizes)
    # back to conv1d friendly (B, C, T) format
    out_streams = map(lambda x: x.transpose(1, 2), out_streams)

    out_syn = []
    for sidx, os in enumerate(out_streams):
        out_stream_syn = torch.zeros_like(os)
        a = analysis_filts[sidx].get_filt_coefs()
        # apply IIR filter for each dimiesion
        for idx in range(os.shape[1]):
            # NOTE: scipy.signal.lfilter accespts b, a in order,
            # but torchaudio expect the oppsite; a, b in order
            ai = a[idx].view(-1).flip(0)
            bi = torch.zeros_like(ai)
            bi[0] = 1
            out_stream_syn[:, idx, :] = lfilter(os[:, idx, :], ai, bi, clamp=False)
        out_syn += [out_stream_syn]

    out_syn = torch.cat(out_syn, 1)
    return out_syn.transpose(1, 2)


class Conv1dResnetSAR(Conv1dResnet):
    """Conv1dResnet with shallow AR structure

    Args:
        in_dim (int): the dimension of the input
        hidden_dim (int): the dimension of the hidden state
        out_dim (int): the dimension of the output
        num_layers (int): the number of layers
        stream_sizes (list): Stream sizes
        ar_orders (list): Filter dimensions for each stream.
        init_type (str): the type of weight initialization
    """

    def __init__(
        self,
        in_dim,
        hidden_dim,
        out_dim,
        num_layers=4,
        stream_sizes=None,
        ar_orders=None,
        init_type="none",
        **kwargs,
    ):
        super().__init__(
            in_dim=in_dim, hidden_dim=hidden_dim, out_dim=out_dim, num_layers=num_layers
        )

        if "dropout" in kwargs:
            warn(
                "dropout argument in Conv1dResnetSAR is deprecated"
                " and will be removed in future versions"
            )

        if stream_sizes is None:
            stream_sizes = [180, 3, 1, 15]
        if ar_orders is None:
            ar_orders = [20, 200, 20, 20]
        self.stream_sizes = stream_sizes

        init_weights(self, init_type)

        self.analysis_filts = nn.ModuleList()
        for s, K in zip(stream_sizes, ar_orders):
            self.analysis_filts += [TrTimeInvFIRFilter(s, K + 1)]

    def preprocess_target(self, y):
        assert sum(self.stream_sizes) == y.shape[-1]
        ys = split_streams(y, self.stream_sizes)
        for idx, yi in enumerate(ys):
            ys[idx] = self.analysis_filts[idx](yi.transpose(1, 2)).transpose(1, 2)
        return torch.cat(ys, -1)

    def inference(self, x, lengths=None, y=None):
        out = self.model(x.transpose(1, 2)).transpose(1, 2)
        return _shallow_ar_inference(out, self.stream_sizes, self.analysis_filts)


class FFN(BaseModel):
    """Feed-forward network

    Args:
        in_dim (int): the dimension of the input
        hidden_dim (int): the dimension of the hidden state
        out_dim (int): the dimension of the output
        num_layers (int): the number of layers
        dropout (float): dropout rate
        init_type (str): the type of weight initialization
        last_sigmoid (bool): whether to apply sigmoid on the output
    """

    def __init__(
        self,
        in_dim,
        hidden_dim,
        out_dim,
        num_layers=2,
        dropout=0.0,
        init_type="none",
        last_sigmoid=False,
    ):
        super(FFN, self).__init__()
        self.first_linear = nn.Linear(in_dim, hidden_dim)
        self.hidden_layers = nn.ModuleList(
            [nn.Linear(hidden_dim, hidden_dim) for _ in range(num_layers)]
        )
        self.last_linear = nn.Linear(hidden_dim, out_dim)
        self.relu = nn.ReLU()
        self.dropout = nn.Dropout(dropout)
        self.last_sigmoid = last_sigmoid
        init_weights(self, init_type)

    def forward(self, x, lengths=None, y=None):
        """Forward step

        Args:
            x (torch.Tensor): the input tensor
            lengths (torch.Tensor): the lengths of the input tensor
            y (torch.Tensor): the optional target tensor

        Returns:
            torch.Tensor: the output tensor
        """
        h = self.relu(self.first_linear(x))
        for hl in self.hidden_layers:
            h = self.dropout(self.relu(hl(h)))
        out = self.last_linear(h)
        out = torch.sigmoid(out) if self.last_sigmoid else out
        return out


# For compatibility
FeedForwardNet = FFN


class LSTMRNN(BaseModel):
    """LSTM-based recurrent neural network

    Args:
        in_dim (int): the dimension of the input
        hidden_dim (int): the dimension of the hidden state
        out_dim (int): the dimension of the output
        num_layers (int): the number of layers
        bidirectional (bool): whether to use bidirectional LSTM
        dropout (float): dropout rate
        init_type (str): the type of weight initialization
    """

    def __init__(
        self,
        in_dim,
        hidden_dim,
        out_dim,
        num_layers=1,
        bidirectional=True,
        dropout=0.0,
        init_type="none",
    ):
        super(LSTMRNN, self).__init__()
        self.hidden_dim = hidden_dim
        self.num_layers = num_layers
        self.num_direction = 2 if bidirectional else 1
        self.lstm = nn.LSTM(
            in_dim,
            hidden_dim,
            num_layers,
            bidirectional=bidirectional,
            batch_first=True,
            dropout=dropout,
        )
        self.hidden2out = nn.Linear(self.num_direction * self.hidden_dim, out_dim)
        init_weights(self, init_type)

    def forward(self, x, lengths, y=None):
        """Forward step

        Args:
            x (torch.Tensor): the input tensor
            lengths (torch.Tensor): the lengths of the input tensor
            y (torch.Tensor): the optional target tensor

        Returns:
            torch.Tensor: the output tensor
        """
        if isinstance(lengths, torch.Tensor):
            lengths = lengths.to("cpu")
        x = pack_padded_sequence(x, lengths, batch_first=True)
        out, _ = self.lstm(x)
        out, _ = pad_packed_sequence(out, batch_first=True)
        out = self.hidden2out(out)
        return out


class LSTMRNNSAR(LSTMRNN):
    """LSTM-RNN with shallow AR structure

    Args:
        in_dim (int): the dimension of the input
        hidden_dim (int): the dimension of the hidden state
        out_dim (int): the dimension of the output
        num_layers (int): the number of layers
        bidirectional (bool): whether to use bidirectional LSTM
        dropout (float): dropout rate
        stream_sizes (list): Stream sizes
        ar_orders (list): Filter dimensions for each stream.
        init_type (str): the type of weight initialization
    """

    def __init__(
        self,
        in_dim,
        hidden_dim,
        out_dim,
        num_layers=1,
        bidirectional=True,
        dropout=0.0,
        stream_sizes=None,
        ar_orders=None,
        init_type="none",
    ):
        super().__init__(
            in_dim, hidden_dim, out_dim, num_layers, bidirectional, dropout, init_type
        )
        if stream_sizes is None:
            stream_sizes = [180, 3, 1, 15]
        if ar_orders is None:
            ar_orders = [20, 200, 20, 20]

        self.stream_sizes = stream_sizes
        self.analysis_filts = nn.ModuleList()
        for s, K in zip(stream_sizes, ar_orders):
            self.analysis_filts += [TrTimeInvFIRFilter(s, K + 1)]

    def preprocess_target(self, y):
        assert sum(self.stream_sizes) == y.shape[-1]
        ys = split_streams(y, self.stream_sizes)
        for idx, yi in enumerate(ys):
            ys[idx] = self.analysis_filts[idx](yi.transpose(1, 2)).transpose(1, 2)
        return torch.cat(ys, -1)

    def inference(self, x, lengths=None, y=None):
        out = self.forward(x, lengths)
        return _shallow_ar_inference(out, self.stream_sizes, self.analysis_filts)


class RMDN(BaseModel):
    """RNN-based mixture density networks (MDN)

    Args:
        in_dim (int): the dimension of the input
        hidden_dim (int): the dimension of the hidden state
        out_dim (int): the dimension of the output
        num_layers (int): the number of layers
        bidirectional (bool): whether to use bidirectional LSTM
        dropout (float): dropout rate
        num_gaussians (int): the number of gaussians
        dim_wise (bool): whether to use dimension-wise or not
        init_type (str): the type of weight initialization
    """

    def __init__(
        self,
        in_dim,
        hidden_dim,
        out_dim,
        num_layers=1,
        bidirectional=True,
        dropout=0.0,
        num_gaussians=8,
        dim_wise=False,
        init_type="none",
    ):
        super(RMDN, self).__init__()
        self.linear = nn.Linear(in_dim, hidden_dim)
        self.relu = nn.ReLU()
        self.num_direction = 2 if bidirectional else 1
        self.lstm = nn.LSTM(
            hidden_dim,
            hidden_dim,
            num_layers,
            bidirectional=bidirectional,
            batch_first=True,
            dropout=dropout,
        )
        self.mdn = MDNLayer(
            in_dim=self.num_direction * hidden_dim,
            out_dim=out_dim,
            num_gaussians=num_gaussians,
            dim_wise=dim_wise,
        )
        init_weights(self, init_type)

    def prediction_type(self):
        return PredictionType.PROBABILISTIC

    def forward(self, x, lengths, y=None):
        """Forward step

        Args:
            x (torch.Tensor): the input tensor
            lengths (torch.Tensor): the lengths of the input tensor
            y (torch.Tensor): the optional target tensor

        Returns:
            torch.Tensor: the output tensor
        """
        if isinstance(lengths, torch.Tensor):
            lengths = lengths.to("cpu")
        out = self.linear(x)
        sequence = pack_padded_sequence(self.relu(out), lengths, batch_first=True)
        out, _ = self.lstm(sequence)
        out, _ = pad_packed_sequence(out, batch_first=True)
        out = self.mdn(out)
        return out

    def inference(self, x, lengths=None):
        """Inference step

        Find the most likely mean and variance

        Args:
            x (torch.Tensor): the input tensor
            lengths (torch.Tensor): the lengths of the input tensor

        Returns:
            tuple: mean and variance of the output features
        """
        log_pi, log_sigma, mu = self.forward(x, lengths)
        sigma, mu = mdn_get_most_probable_sigma_and_mu(log_pi, log_sigma, mu)
        return mu, sigma


class MDN(BaseModel):
    """Mixture density networks (MDN) with FFN

    .. warning::

        It is recommended to use MDNv2 instead, unless you want to
        fine-turn from a old checkpoint of MDN.

    Args:
        in_dim (int): the dimension of the input
        hidden_dim (int): the dimension of the hidden state
        out_dim (int): the dimension of the output
        num_layers (int): the number of layers
        num_gaussians (int): the number of gaussians
        dim_wise (bool): whether to use dimension-wise or not
        init_type (str): the type of weight initialization
    """

    def __init__(
        self,
        in_dim,
        hidden_dim,
        out_dim,
        num_layers=1,
        num_gaussians=8,
        dim_wise=False,
        init_type="none",
        **kwargs,
    ):
        super(MDN, self).__init__()
        if "dropout" in kwargs:
            warn(
                "dropout argument in MDN is deprecated"
                " and will be removed in future versions"
            )
        model = [nn.Linear(in_dim, hidden_dim), nn.ReLU()]
        if num_layers > 1:
            for _ in range(num_layers - 1):
                model += [nn.Linear(hidden_dim, hidden_dim), nn.ReLU()]
        model += [
            MDNLayer(
                in_dim=hidden_dim,
                out_dim=out_dim,
                num_gaussians=num_gaussians,
                dim_wise=dim_wise,
            )
        ]
        self.model = nn.Sequential(*model)
        init_weights(self, init_type)

    def prediction_type(self):
        return PredictionType.PROBABILISTIC

    def forward(self, x, lengths=None, y=None):
        """Forward step

        Args:
            x (torch.Tensor): the input tensor
            lengths (torch.Tensor): the lengths of the input tensor
            y (torch.Tensor): the optional target tensor

        Returns:
            torch.Tensor: the output tensor
        """
        return self.model(x)

    def inference(self, x, lengths=None):
        """Inference step

        Find the most likely mean and variance

        Args:
            x (torch.Tensor): the input tensor
            lengths (torch.Tensor): the lengths of the input tensor

        Returns:
            tuple: mean and variance of the output features
        """
        log_pi, log_sigma, mu = self.forward(x, lengths)
        sigma, mu = mdn_get_most_probable_sigma_and_mu(log_pi, log_sigma, mu)
        return mu, sigma


class MDNv2(BaseModel):
    """Mixture density networks (MDN) with FFN

    MDN (v1) + Dropout

    Args:
        in_dim (int): the dimension of the input
        hidden_dim (int): the dimension of the hidden state
        out_dim (int): the dimension of the output
        num_layers (int): the number of layers
        dropout (float): dropout rate
        num_gaussians (int): the number of gaussians
        dim_wise (bool): whether to use dimension-wise or not
        init_type (str): the type of weight initialization
    """

    def __init__(
        self,
        in_dim,
        hidden_dim,
        out_dim,
        num_layers=1,
        dropout=0.5,
        num_gaussians=8,
        dim_wise=False,
        init_type="none",
    ):
        super(MDNv2, self).__init__()
        model = [nn.Linear(in_dim, hidden_dim), nn.ReLU(), nn.Dropout(dropout)]
        if num_layers > 1:
            for _ in range(num_layers - 1):
                model += [
                    nn.Linear(hidden_dim, hidden_dim),
                    nn.ReLU(),
                    nn.Dropout(dropout),
                ]
        model += [
            MDNLayer(
                in_dim=hidden_dim,
                out_dim=out_dim,
                num_gaussians=num_gaussians,
                dim_wise=dim_wise,
            )
        ]
        self.model = nn.Sequential(*model)
        init_weights(self, init_type)

    def prediction_type(self):
        return PredictionType.PROBABILISTIC

    def forward(self, x, lengths=None, y=None):
        """Forward step

        Args:
            x (torch.Tensor): the input tensor
            lengths (torch.Tensor): the lengths of the input tensor
            y (torch.Tensor): the optional target tensor

        Returns:
            torch.Tensor: the output tensor
        """
        return self.model(x)

    def inference(self, x, lengths=None):
        """Inference step

        Find the most likely mean and variance

        Args:
            x (torch.Tensor): the input tensor
            lengths (torch.Tensor): the lengths of the input tensor

        Returns:
            tuple: mean and variance of the output features
        """
        log_pi, log_sigma, mu = self.forward(x, lengths)
        sigma, mu = mdn_get_most_probable_sigma_and_mu(log_pi, log_sigma, mu)
        return mu, sigma


class Conv1dResnetMDN(BaseModel):
    """Conv1dResnet with MDN output layer"""

    def __init__(
        self,
        in_dim,
        hidden_dim,
        out_dim,
        num_layers=4,
        num_gaussians=8,
        dim_wise=False,
        init_type="none",
        **kwargs,
    ):
        super().__init__()

        if "dropout" in kwargs:
            warn("dropout argument in Conv1dResnet is deprecated")

        model = [
            Conv1dResnet(
                in_dim=in_dim,
                hidden_dim=hidden_dim,
                out_dim=hidden_dim,
                num_layers=num_layers,
            ),
            nn.ReLU(),
            MDNLayer(
                in_dim=hidden_dim,
                out_dim=out_dim,
                num_gaussians=num_gaussians,
                dim_wise=dim_wise,
            ),
        ]
        self.model = nn.Sequential(*model)
        init_weights(self, init_type)

    def prediction_type(self):
        return PredictionType.PROBABILISTIC

    def forward(self, x, lengths=None, y=None):
        """Forward step

        Args:
            x (torch.Tensor): the input tensor
            lengths (torch.Tensor): the lengths of the input tensor
            y (torch.Tensor): the optional target tensor

        Returns:
            torch.Tensor: the output tensor
        """
        return self.model(x)

    def inference(self, x, lengths=None):
        """Inference step

        Find the most likely mean and variance

        Args:
            x (torch.Tensor): the input tensor
            lengths (torch.Tensor): the lengths of the input tensor

        Returns:
            tuple: mean and variance of the output features
        """
        log_pi, log_sigma, mu = self.forward(x, lengths)
        sigma, mu = mdn_get_most_probable_sigma_and_mu(log_pi, log_sigma, mu)
        return mu, sigma


class FFConvLSTM(BaseModel):
    """FFN + Conv1d + LSTM

    A model proposed in :cite:t:`hono2021sinsy` without residual F0 prediction.

    Args:
        in_dim (int): the dimension of the input
        ff_hidden_dim (int): the dimension of the hidden state of the FFN
        conv_hidden_dim (int): the dimension of the hidden state of the conv1d
        lstm_hidden_dim (int): the dimension of the hidden state of the LSTM
        out_dim (int): the dimension of the output
        dropout (float): dropout rate
        num_lstm_layers (int): the number of layers of the LSTM
        bidirectional (bool): whether to use bidirectional LSTM
        init_type (str): the type of weight initialization
        use_mdn (bool): whether to use MDN or not
        dim_wise (bool): whether to use dimension-wise or not
        num_gaussians (int): the number of gaussians
        in_ph_start_idx (int): the start index of phoneme identity in a hed file
        in_ph_end_idx (int): the end index of phoneme identity in a hed file
        embed_dim (int): the dimension of the phoneme embedding
    """

    def __init__(
        self,
        in_dim,
        ff_hidden_dim=2048,
        conv_hidden_dim=1024,
        lstm_hidden_dim=256,
        out_dim=67,
        dropout=0.0,
        num_lstm_layers=2,
        bidirectional=True,
        init_type="none",
        use_mdn=False,
        dim_wise=True,
        num_gaussians=4,
        in_ph_start_idx: int = 1,
        in_ph_end_idx: int = 50,
        embed_dim=None,
        enforce_sorted=True,
    ):
        super().__init__()
        self.in_dim = in_dim
        self.out_dim = out_dim
        self.in_ph_start_idx = in_ph_start_idx
        self.in_ph_end_idx = in_ph_end_idx
        self.num_vocab = in_ph_end_idx - in_ph_start_idx
        self.embed_dim = embed_dim
        self.use_mdn = use_mdn
        self.enforce_sorted = enforce_sorted

        if self.embed_dim is not None:
            assert in_dim > self.num_vocab
            self.emb = nn.Embedding(self.num_vocab, embed_dim)
            self.fc_in = nn.Linear(in_dim - self.num_vocab, embed_dim)
            ff_in_dim = embed_dim
        else:
            ff_in_dim = in_dim

        self.ff = nn.Sequential(
            nn.Linear(ff_in_dim, ff_hidden_dim),
            nn.ReLU(),
            nn.Linear(ff_hidden_dim, ff_hidden_dim),
            nn.ReLU(),
            nn.Linear(ff_hidden_dim, ff_hidden_dim),
            nn.ReLU(),
        )

        self.conv = nn.Sequential(
            nn.ReflectionPad1d(3),
            nn.Conv1d(ff_hidden_dim, conv_hidden_dim, kernel_size=7, padding=0),
            nn.BatchNorm1d(conv_hidden_dim),
            nn.ReLU(),
            nn.ReflectionPad1d(3),
            nn.Conv1d(conv_hidden_dim, conv_hidden_dim, kernel_size=7, padding=0),
            nn.BatchNorm1d(conv_hidden_dim),
            nn.ReLU(),
            nn.ReflectionPad1d(3),
            nn.Conv1d(conv_hidden_dim, conv_hidden_dim, kernel_size=7, padding=0),
            nn.BatchNorm1d(conv_hidden_dim),
            nn.ReLU(),
        )

        num_direction = 2 if bidirectional else 1
        self.lstm = nn.LSTM(
            conv_hidden_dim,
            lstm_hidden_dim,
            num_lstm_layers,
            bidirectional=True,
            batch_first=True,
            dropout=dropout,
        )

        last_in_dim = num_direction * lstm_hidden_dim
        if self.use_mdn:
            assert dim_wise
            self.fc = MDNLayer(
                in_dim=last_in_dim,
                out_dim=out_dim,
                num_gaussians=num_gaussians,
                dim_wise=dim_wise,
            )
        else:
            self.fc = nn.Linear(last_in_dim, out_dim)
        init_weights(self, init_type)

    def prediction_type(self):
        return (
            PredictionType.PROBABILISTIC
            if self.use_mdn
            else PredictionType.DETERMINISTIC
        )

    def forward(self, x, lengths=None, y=None):
        if isinstance(lengths, torch.Tensor):
            lengths = lengths.to("cpu")

        if self.embed_dim is not None:
            x_first, x_ph_onehot, x_last = torch.split(
                x,
                [
                    self.in_ph_start_idx,
                    self.num_vocab,
                    self.in_dim - self.num_vocab - self.in_ph_start_idx,
                ],
                dim=-1,
            )
            x_ph = torch.argmax(x_ph_onehot, dim=-1)
            # Make sure to have one-hot vector
            assert (x_ph_onehot.sum(-1) <= 1).all()
            x = self.emb(x_ph) + self.fc_in(torch.cat([x_first, x_last], dim=-1))

        out = self.ff(x)
        out = self.conv(out.transpose(1, 2)).transpose(1, 2)
        sequence = pack_padded_sequence(
            out, lengths, batch_first=True, enforce_sorted=self.enforce_sorted
        )
        out, _ = self.lstm(sequence)
        out, _ = pad_packed_sequence(out, batch_first=True, total_length=x.size(1))

        return self.fc(out)

    def inference(self, x, lengths=None):
        if self.use_mdn:
            log_pi, log_sigma, mu = self(x, lengths)
            sigma, mu = mdn_get_most_probable_sigma_and_mu(log_pi, log_sigma, mu)
            return mu, sigma
        else:
            return self(x, lengths)


class VariancePredictor(BaseModel):
    """Variance predictor in :cite:t:`ren2020fastspeech`.

    The model is composed of stacks of Conv1d + ReLU + LayerNorm layers.
    The model can be used for duration or pitch prediction.

    Args:
        in_dim (int): the input dimension
        out_dim (int): the output dimension
        num_layers (int): the number of layers
        hidden_dim (int): the hidden dimension
        kernel_size (int): the kernel size
        dropout (float): the dropout rate
        init_type (str): the initialization type
        use_mdn (bool): whether to use MDN or not
        num_gaussians (int): the number of gaussians
        dim_wise (bool): whether to use dim-wise or not
        in_ph_start_idx (int): the start index of phoneme identity in a hed file
        in_ph_end_idx (int): the end index of phoneme identity in a hed file
        embed_dim (int): the dimension of the phoneme embedding
        mask_indices (list): the input feature indices to be masked.
            e.g., specify pitch_idx to mask pitch features.
    """

    def __init__(
        self,
        in_dim,
        out_dim,
        num_layers=5,
        hidden_dim=256,
        kernel_size=5,
        dropout=0.5,
        init_type="none",
        use_mdn=False,
        num_gaussians=1,
        dim_wise=False,
        in_ph_start_idx: int = 1,
        in_ph_end_idx: int = 50,
        embed_dim=None,
        mask_indices=None,
    ):
        super().__init__()
        self.in_dim = in_dim
        self.out_dim = out_dim
        self.use_mdn = use_mdn
        self.in_ph_start_idx = in_ph_start_idx
        self.in_ph_end_idx = in_ph_end_idx
        self.num_vocab = in_ph_end_idx - in_ph_start_idx
        self.embed_dim = embed_dim
        self.use_mdn = use_mdn
        self.mask_indices = mask_indices

        if self.embed_dim is not None:
            assert in_dim > self.num_vocab
            self.emb = nn.Embedding(self.num_vocab, embed_dim)
            self.fc_in = nn.Linear(in_dim - self.num_vocab, embed_dim)
            in_dim = embed_dim

        conv = nn.ModuleList()
        for idx in range(num_layers):
            in_channels = in_dim if idx == 0 else hidden_dim
            conv += [
                nn.Sequential(
                    nn.Conv1d(
                        in_channels,
                        hidden_dim,
                        kernel_size,
                        stride=1,
                        padding=(kernel_size - 1) // 2,
                    ),
                    nn.ReLU(),
                    LayerNorm(hidden_dim, dim=1),
                    nn.Dropout(dropout),
                )
            ]
        self.conv = nn.Sequential(*conv)
        if self.use_mdn:
            self.mdn_layer = MDNLayer(
                hidden_dim, out_dim, num_gaussians=num_gaussians, dim_wise=dim_wise
            )
        else:
            self.fc = nn.Linear(hidden_dim, out_dim)

        init_weights(self, init_type)

    def prediction_type(self):
        return (
            PredictionType.PROBABILISTIC
            if self.use_mdn
            else PredictionType.DETERMINISTIC
        )

    def forward(self, x, lengths=None, y=None):
        """Forward step

        Args:
            x (torch.Tensor): the input tensor
            lengths (torch.Tensor): the lengths of the input tensor
            y (torch.Tensor): the optional target tensor

        Returns:
            torch.Tensor: the output tensor
        """
        # Masking specified features
        if self.mask_indices is not None:
            for idx in self.mask_indices:
                x[:, :, idx] *= 0.0

        if self.embed_dim is not None:
            x_first, x_ph_onehot, x_last = torch.split(
                x,
                [
                    self.in_ph_start_idx,
                    self.num_vocab,
                    self.in_dim - self.num_vocab - self.in_ph_start_idx,
                ],
                dim=-1,
            )
            x_ph = torch.argmax(x_ph_onehot, dim=-1)
            # Make sure to have one-hot vector
            assert (x_ph_onehot.sum(-1) <= 1).all()
            x = self.emb(x_ph) + self.fc_in(torch.cat([x_first, x_last], dim=-1))

        out = self.conv(x.transpose(1, 2)).transpose(1, 2)

        if self.use_mdn:
            return self.mdn_layer(out)
        else:
            return self.fc(out)

    def inference(self, x, lengths=None):
        """Inference step

        Find the most likely mean and variance if use_mdn is True

        Args:
            x (torch.Tensor): the input tensor
            lengths (torch.Tensor): the lengths of the input tensor

        Returns:
            tuple: mean and variance of the output features
        """
        if self.use_mdn:
            log_pi, log_sigma, mu = self(x, lengths)
            sigma, mu = mdn_get_most_probable_sigma_and_mu(log_pi, log_sigma, mu)
            return mu, sigma
        else:
            return self(x, lengths)


class LSTMEncoder(BaseModel):
    """LSTM encoder

    A simple LSTM-based encoder

    Args:
        in_dim (int): the input dimension
        hidden_dim (int): the hidden dimension
        out_dim (int): the output dimension
        num_layers (int): the number of layers
        bidirectional (bool): whether to use bidirectional or not
        dropout (float): the dropout rate
        init_type (str): the initialization type
        in_ph_start_idx (int): the start index of phonetic context in a hed file
        in_ph_end_idx (int): the end index of phonetic context in a hed file
        embed_dim (int): the embedding dimension
    """

    def __init__(
        self,
        in_dim: int,
        hidden_dim: int,
        out_dim: int,
        num_layers: int = 1,
        bidirectional: bool = True,
        dropout: float = 0.0,
        init_type: str = "none",
        in_ph_start_idx: int = 1,
        in_ph_end_idx: int = 50,
        embed_dim=None,
        enforce_sorted=True,
    ):
        super(LSTMEncoder, self).__init__()
        self.in_dim = in_dim
        self.in_ph_start_idx = in_ph_start_idx
        self.in_ph_end_idx = in_ph_end_idx
        self.num_vocab = in_ph_end_idx - in_ph_start_idx
        self.embed_dim = embed_dim
        self.enforce_sorted = enforce_sorted

        if self.embed_dim is not None:
            assert in_dim > self.num_vocab
            self.emb = nn.Embedding(self.num_vocab, embed_dim)
            self.fc_in = nn.Linear(in_dim - self.num_vocab, embed_dim)
            lstm_in_dim = embed_dim
        else:
            lstm_in_dim = in_dim

        self.num_layers = num_layers
        num_direction = 2 if bidirectional else 1
        self.lstm = nn.LSTM(
            lstm_in_dim,
            hidden_dim,
            num_layers,
            bidirectional=bidirectional,
            batch_first=True,
            dropout=dropout,
        )
        self.hidden2out = nn.Linear(num_direction * hidden_dim, out_dim)
        init_weights(self, init_type)

    def forward(self, x, lengths, y=None):
        if self.embed_dim is not None:
            x_first, x_ph_onehot, x_last = torch.split(
                x,
                [
                    self.in_ph_start_idx,
                    self.num_vocab,
                    self.in_dim - self.num_vocab - self.in_ph_start_idx,
                ],
                dim=-1,
            )
            x_ph = torch.argmax(x_ph_onehot, dim=-1)
            # Make sure to have one-hot vector
            assert (x_ph_onehot.sum(-1) <= 1).all()
            x = self.emb(x_ph) + self.fc_in(torch.cat([x_first, x_last], dim=-1))

        if isinstance(lengths, torch.Tensor):
            lengths = lengths.to("cpu")
        total_length = x.size(1)
        x = pack_padded_sequence(
            x, lengths, batch_first=True, enforce_sorted=self.enforce_sorted
        )
        out, _ = self.lstm(x)
        out, _ = pad_packed_sequence(out, batch_first=True, total_length=total_length)
        out = self.hidden2out(out)
        return out


class TransformerEncoder(BaseModel):
    """Transformer encoder


    .. warning::

        So far this is not well tested. Maybe be removed in the future.

    Args:
        in_dim (int): the input dimension
        out_dim (int): the output dimension
        hidden_dim (int): the hidden dimension
        attention_dim (int): the attention dimension
        num_heads (int): the number of heads
        num_layers (int): the number of layers
        kernel_size (int): the kernel size
        dropout (float): the dropout rate
        reduction_factor (int): the reduction factor
        init_type (str): the initialization type
        downsample_by_conv (bool): whether to use convolutional downsampling or not
        in_ph_start_idx (int): the start index of phonetic context in a hed file
        in_ph_end_idx (int): the end index of phonetic context in a hed file
        embed_dim (int): the embedding dimension
    """

    def __init__(
        self,
        in_dim,
        out_dim,
        hidden_dim,
        attention_dim,
        num_heads=2,
        num_layers=2,
        kernel_size=3,
        dropout=0.1,
        reduction_factor=1,
        init_type="none",
        downsample_by_conv=False,
        in_ph_start_idx: int = 1,
        in_ph_end_idx: int = 50,
        embed_dim=None,
    ):
        super().__init__()
        self.in_dim = in_dim
        self.out_dim = out_dim
        self.in_ph_start_idx = in_ph_start_idx
        self.in_ph_end_idx = in_ph_end_idx
        self.num_vocab = in_ph_end_idx - in_ph_start_idx
        self.embed_dim = embed_dim

        if self.embed_dim is not None:
            assert in_dim > self.num_vocab
            self.emb = nn.Embedding(self.num_vocab, embed_dim)
            self.fc_in = nn.Linear(in_dim - self.num_vocab, embed_dim)
            self.fc = nn.Linear(embed_dim, hidden_dim)
        else:
            self.emb = None
            self.fc_in = None
            self.fc = nn.Linear(in_dim, hidden_dim)
        self.reduction_factor = reduction_factor
        self.encoder = _TransformerEncoder(
            hidden_channels=hidden_dim,
            filter_channels=attention_dim,
            n_heads=num_heads,
            n_layers=num_layers,
            kernel_size=kernel_size,
            p_dropout=dropout,
        )
        self.fc_out = nn.Linear(hidden_dim, out_dim * reduction_factor)

        if reduction_factor > 1 and downsample_by_conv:
            self.conv_downsample = nn.Conv1d(
                in_dim,
                in_dim,
                kernel_size=reduction_factor,
                stride=reduction_factor,
                groups=in_dim,
            )
        else:
            self.conv_downsample = None

        for f in [self.fc_in, self.emb, self.fc, self.fc_out]:
            if f is not None:
                init_weights(f, init_type)

    def forward(self, x, lengths=None, y=None):
        """Forward pass

        Args:
            x (torch.Tensor): input tensor
            lengths (torch.Tensor): input sequence lengths
            y (torch.Tensor): target tensor (optional)

        Returns:
            torch.Tensor: output tensor
        """
        if isinstance(lengths, list):
            lengths = torch.tensor(lengths).to(x.device)

        if self.embed_dim is not None:
            x_first, x_ph_onehot, x_last = torch.split(
                x,
                [
                    self.in_ph_start_idx,
                    self.num_vocab,
                    self.in_dim - self.num_vocab - self.in_ph_start_idx,
                ],
                dim=-1,
            )
            x_ph = torch.argmax(x_ph_onehot, dim=-1)
            # Make sure to have one-hot vector
            assert (x_ph_onehot.sum(-1) <= 1).all()
            x = self.emb(x_ph) + self.fc_in(torch.cat([x_first, x_last], dim=-1))

        # Adjust lengths based on the reduction factor
        if self.reduction_factor > 1:
            lengths = (lengths / self.reduction_factor).long()
            if self.conv_downsample is not None:
                x = self.conv_downsample(x.transpose(1, 2)).transpose(1, 2)
            else:
                x = x[:, self.reduction_factor - 1 :: self.reduction_factor]

        x = self.fc(x)
        # (B, T, C) -> (B, C, T)
        x = x.transpose(1, 2)
        x_mask = sequence_mask(lengths, x.shape[2]).unsqueeze(1).to(x.device)
        x = self.encoder(x * x_mask, x_mask)

        # (B, C, T) -> (B, T, C)
        x = self.fc_out(x.transpose(1, 2)).view(x.shape[0], -1, self.out_dim)

        return x


# For backward compatibility


def ResF0Conv1dResnet(*args, **kwargs):
    from nnsvs.acoustic_models import ResF0Conv1dResnet

    return ResF0Conv1dResnet(*args, **kwargs)


def ResF0Conv1dResnetMDN(*args, **kwargs):
    from nnsvs.acoustic_models import ResF0Conv1dResnetMDN

    return ResF0Conv1dResnetMDN(*args, **kwargs)


def ResF0VariancePredictor(*args, **kwargs):
    from nnsvs.acoustic_models import ResF0VariancePredictor

    return ResF0VariancePredictor(*args, **kwargs)


def ResSkipF0FFConvLSTM(*args, **kwargs):
    from nnsvs.acoustic_models import ResSkipF0FFConvLSTM

    return ResSkipF0FFConvLSTM(*args, **kwargs)



# -*- coding: utf-8 -*-

# Copyright 2020 Yi-Chiao Wu (Nagoya University)
# based on a WaveNet script by Tomoki Hayashi (Nagoya University)
# (https://github.com/kan-bayashi/PytorchWaveNetVocoder)
# based on sprocket-vc script by Kazuhiro Kobayashi (Nagoya University)
# (https://github.com/k2kobayashi/sprocket)
#  MIT License (https://opensource.org/licenses/MIT)

"""Filters."""

import numpy as np
from scipy.signal import firwin, lfilter

NUMTAPS = 255


def low_cut_filter(x, fs, cutoff=70):
    """Low-cut filter

    Args:
        x (ndarray): Waveform sequence
        fs (int): Sampling frequency
        cutoff (float): Cutoff frequency of low cut filter

    Return:
        (ndarray): Low cut filtered waveform sequence

    """
    nyquist = fs // 2
    norm_cutoff = cutoff / nyquist
    numtaps = NUMTAPS
    fil = firwin(numtaps, norm_cutoff, pass_zero=False)
    lcf_x = lfilter(fil, 1, x)

    return lcf_x


def low_pass_filter(x, fs, cutoff=70):
    """Low-pass filter

    Args:
        x (ndarray): Waveform sequence
        fs (int): Sampling frequency
        cutoff (float): Cutoff frequency of low pass filter

    Return:
        (ndarray): Low pass filtered waveform sequence

    """
    nyquist = fs // 2
    norm_cutoff = cutoff / nyquist
    numtaps = NUMTAPS
    fil = firwin(numtaps, norm_cutoff, pass_zero=True)
    x_pad = np.pad(x, (numtaps, numtaps), "edge")
    lpf_x = lfilter(fil, 1, x_pad)
    lpf_x = lpf_x[numtaps + numtaps // 2 : -numtaps // 2]

    return lpf_x



from os.path import dirname, join

import hydra
import torch
from nnsvs.model import MDN
from omegaconf import OmegaConf


# https://github.com/r9y9/nnsvs/pull/114#issuecomment-1156631058
def test_mdn_compat():
    config = OmegaConf.load(join(dirname(__file__), "data", "mdn_test.yaml"))
    model = hydra.utils.instantiate(config.netG)
    checkpoint = torch.load(join(dirname(__file__), "data", "mdn_test.pth"))
    model.load_state_dict(checkpoint["state_dict"])
    assert isinstance(model, MDN)



from os.path import exists

import hydra
import joblib
import numpy as np
from hydra.utils import to_absolute_path
from nnsvs.logger import getLogger
from omegaconf import DictConfig, OmegaConf
from sklearn.preprocessing import MinMaxScaler, StandardScaler


@hydra.main(config_path="conf/fit_scaler", config_name="config")
def my_app(config: DictConfig) -> None:
    logger = getLogger(config.verbose)
    logger.info(OmegaConf.to_yaml(config))

    list_path = to_absolute_path(config.list_path)
    out_path = to_absolute_path(config.out_path)

    if config.external_scaler is not None:
        scaler = joblib.load(config.external_scaler)
    else:
        scaler = hydra.utils.instantiate(config.scaler)
    with open(list_path) as f:
        for path in f:
            path = to_absolute_path(path.strip())
            c = np.load(path)
            if "out_acoustic/" in path:
                assert "org" in path
                in_path = path.replace("out_acoustic/", "in_acoustic/")
                in_feats = np.load(in_path)
                note_frame_indices = in_feats[:, config.in_rest_idx] <= 0
                assert exists(in_path)
                # Removed non-voice segments to compute statistics
                c = c[note_frame_indices]
            scaler.partial_fit(c)
        joblib.dump(scaler, out_path)

    if config.verbose > 0:
        if isinstance(scaler, StandardScaler):
            logger.info("mean:\n%s", scaler.mean_)
            logger.info("std:\n%s", np.sqrt(scaler.var_))
        if isinstance(scaler, MinMaxScaler):
            logger.info("data min:\n%s", scaler.data_min_)
            logger.info("data max:\n%s", scaler.data_max_)


def entry():
    my_app()  # pylint: disable=no-value-for-parameter


if __name__ == "__main__":
    my_app()  # pylint: disable=no-value-for-parameter



# -*- coding: utf-8 -*-

# Copyright 2022 Reo Yoneyama (Nagoya University)
#  MIT License (https://opensource.org/licenses/MIT)

"""Spectral envelopes estimation module based on CheapTrick.

References:
    - https://www.sciencedirect.com/science/article/pii/S0167639314000697
    - https://github.com/mmorise/World

"""

import math

import torch
import torch.fft
import torch.nn as nn


class AdaptiveWindowing(nn.Module):
    """CheapTrick F0 adptive windowing module."""

    def __init__(
        self,
        sampling_rate,
        hop_size,
        fft_size,
        f0_floor,
        f0_ceil,
    ):
        """Initialize AdaptiveWindowing module.

        Args:
            sampling_rate (int): Sampling rate.
            hop_size (int): Hop size.
            fft_size (int): FFT size.
            f0_floor (int): Minimum value of F0.
            f0_ceil (int): Maximum value of F0.

        """
        super(AdaptiveWindowing, self).__init__()

        self.sampling_rate = sampling_rate
        self.hop_size = hop_size
        self.fft_size = fft_size
        self.register_buffer("window", torch.zeros((f0_ceil + 1, fft_size)))
        self.zero_padding = nn.ConstantPad2d((fft_size // 2, fft_size // 2, 0, 0), 0)

        # Pre-calculation of the window functions
        for f0 in range(f0_floor, f0_ceil + 1):
            half_win_len = round(1.5 * self.sampling_rate / f0)
            base_index = torch.arange(
                -half_win_len, half_win_len + 1, dtype=torch.int64
            )
            position = base_index / 1.5 / self.sampling_rate
            left = fft_size // 2 - half_win_len
            right = fft_size // 2 + half_win_len + 1
            window = torch.zeros(fft_size)
            window[left:right] = 0.5 * torch.cos(math.pi * position * f0) + 0.5
            average = torch.sum(window * window).pow(0.5)
            self.window[f0] = window / average

    def forward(self, x, f, power=False):
        """Calculate forward propagation.

        Args:
            x (Tensor): Waveform (B, fft_size // 2 + 1, T).
            f (Tensor): F0 sequence (B, T').
            power (boot): Whether to use power or magnitude.

        Returns:
            Tensor: Power spectrogram (B, bin_size, T').

        """
        # Get the matrix of window functions corresponding to F0
        x = self.zero_padding(x).unfold(1, self.fft_size, self.hop_size)
        windows = self.window[f]

        # Adaptive windowing and calculate power spectrogram.
        # In test, change x[:, : -1, :] to x.
        x = torch.abs(torch.fft.rfft(x[:, :-1, :] * windows))
        x = x.pow(2) if power else x

        return x


class AdaptiveLiftering(nn.Module):
    """CheapTrick F0 adptive windowing module."""

    def __init__(
        self,
        sampling_rate,
        fft_size,
        f0_floor,
        f0_ceil,
        q1=-0.15,
    ):
        """Initialize AdaptiveLiftering module.

        Args:
            sampling_rate (int): Sampling rate.
            fft_size (int): FFT size.
            f0_floor (int): Minimum value of F0.
            f0_ceil (int): Maximum value of F0.
            q1 (float): Parameter to remove effect of adjacent harmonics.

        """
        super(AdaptiveLiftering, self).__init__()

        self.sampling_rate = sampling_rate
        self.bin_size = fft_size // 2 + 1
        self.q1 = q1
        self.q0 = 1.0 - 2.0 * q1
        self.register_buffer(
            "smoothing_lifter", torch.zeros((f0_ceil + 1, self.bin_size))
        )
        self.register_buffer(
            "compensation_lifter", torch.zeros((f0_ceil + 1, self.bin_size))
        )

        # Pre-calculation of the smoothing lifters and compensation lifters
        for f0 in range(f0_floor, f0_ceil + 1):
            smoothing_lifter = torch.zeros(self.bin_size)
            compensation_lifter = torch.zeros(self.bin_size)
            quefrency = torch.arange(1, self.bin_size) / sampling_rate
            smoothing_lifter[0] = 1.0
            smoothing_lifter[1:] = torch.sin(math.pi * f0 * quefrency) / (
                math.pi * f0 * quefrency
            )
            compensation_lifter[0] = self.q0 + 2.0 * self.q1
            compensation_lifter[1:] = self.q0 + 2.0 * self.q1 * torch.cos(
                2.0 * math.pi * f0 * quefrency
            )
            self.smoothing_lifter[f0] = smoothing_lifter
            self.compensation_lifter[f0] = compensation_lifter

    def forward(self, x, f, elim_0th=False):
        """Calculate forward propagation.

        Args:
            x (Tensor): Power spectrogram (B, bin_size, T').
            f (Tensor): F0 sequence (B, T').
            elim_0th (bool): Whether to eliminate cepstram 0th component.

        Returns:
            Tensor: Estimated spectral envelope (B, bin_size, T').

        """
        # Setting the smoothing lifter and compensation lifter
        smoothing_lifter = self.smoothing_lifter[f]
        compensation_lifter = self.compensation_lifter[f]

        # Calculating cepstrum
        tmp = torch.cat((x, torch.flip(x[:, :, 1:-1], [2])), dim=2)
        cepstrum = torch.fft.rfft(torch.log(torch.clamp(tmp, min=1e-7))).real

        # Set the 0th cepstrum to 0
        if elim_0th:
            cepstrum[..., 0] = 0

        # Liftering cepstrum with the lifters
        liftered_cepstrum = cepstrum * smoothing_lifter * compensation_lifter

        # Return the result to the spectral domain
        x = torch.fft.irfft(liftered_cepstrum)[:, :, : self.bin_size]

        return x


class CheapTrick(nn.Module):
    """CheapTrick based spectral envelope estimation module."""

    def __init__(
        self,
        sampling_rate,
        hop_size,
        fft_size,
        f0_floor=70,
        f0_ceil=340,
        uv_threshold=0,
        q1=-0.15,
    ):
        """Initialize AdaptiveLiftering module.

        Args:
            sampling_rate (int): Sampling rate.
            hop_size (int): Hop size.
            fft_size (int): FFT size.
            f0_floor (int): Minimum value of F0.
            f0_ceil (int): Maximum value of F0.
            uv_threshold (float): V/UV determining threshold.
            q1 (float): Parameter to remove effect of adjacent harmonics.

        """
        super(CheapTrick, self).__init__()

        # fft_size must be larger than 3.0 * sampling_rate / f0_floor
        assert fft_size > 3.0 * sampling_rate / f0_floor
        self.f0_floor = f0_floor
        self.f0_ceil = f0_ceil
        self.uv_threshold = uv_threshold

        self.ada_wind = AdaptiveWindowing(
            sampling_rate,
            hop_size,
            fft_size,
            f0_floor,
            f0_ceil,
        )
        self.ada_lift = AdaptiveLiftering(
            sampling_rate,
            fft_size,
            f0_floor,
            f0_ceil,
            q1,
        )

    def forward(self, x, f, power=False, elim_0th=False):
        """Calculate forward propagation.

        Args:
            x (Tensor): Power spectrogram (B, T).
            f (Tensor): F0 sequence (B, T').
            power (boot): Whether to use power or magnitude spectrogram.
            elim_0th (bool): Whether to eliminate cepstram 0th component.

        Returns:
            Tensor: Estimated spectral envelope (B, bin_size, T').

        """
        # Step0: Round F0 values to integers.
        voiced = (f > self.uv_threshold) * torch.ones_like(f)
        f = voiced * f + (1.0 - voiced) * self.f0_ceil
        f = torch.round(torch.clamp(f, min=self.f0_floor, max=self.f0_ceil)).to(
            torch.int64
        )

        # Step1: Adaptive windowing and calculate power or amplitude spectrogram.
        x = self.ada_wind(x, f, power)

        # Step3: Smoothing (log axis) and spectral recovery on the cepstrum domain.
        x = self.ada_lift(x, f, elim_0th)

        return x



from enum import Enum

from torch import nn


class PredictionType(Enum):
    """Prediction types"""

    DETERMINISTIC = 1
    """Deterministic prediction

    Non-MDN single-stream models should use this type.

    Pseudo code:

    .. code-block::

        # training
        y = model(x)
        # inference
        y = model.inference(x)
    """

    PROBABILISTIC = 2
    """Probabilistic prediction with mixture density networks

    MDN-based models should use this type.

    Pseudo code:

    .. code-block::

        # training
        mdn_params = model(x)
        # inference
        mu, sigma = model.inference(x)
    """

    MULTISTREAM_HYBRID = 3
    """Multi-stream preodictions where each prediction can be
    detereministic or probabilistic

    Multi-stream models should use this type.

    Pseudo code:

    .. code-block::

        # training
        feature_streams = model(x) # e.g. (mgc, lf0, vuv, bap) or (mel, lf0, vuv)
        # inference
        y = model.inference(x)

    Note that concatenated features are assumed to be returned during inference.
    """

    DIFFUSION = 4
    """Diffusion model's prediction

    NOTE: may subject to change in the future

    Pseudo code:

    .. code-block::

        # training
        noise, x_recon = model(x)

        # inference
        y = model.inference(x)
    """


class BaseModel(nn.Module):
    """Base class for all models

    If you want to implement your custom model, you should inherit from this class.
    You must need to implement the forward method. Other methods are optional.
    """

    def forward(self, x, lengths=None, y=None):
        """Forward pass

        Args:
            x (tensor): input features
            lengths (tensor): lengths of the input features
            y (tensor): optional target features

        Returns:
            tensor: output features
        """
        pass

    def inference(self, x, lengths=None):
        """Inference method

        If you want to implement custom inference method such as autoregressive sampling,
        please override this method.

        Defaults to call the forward method.

        Args:
            x (tensor): input features
            lengths (tensor): lengths of the input features

        Returns:
            tensor: output features
        """
        return self(x, lengths)

    def preprocess_target(self, y):
        """Preprocess target signals at training time

        This is useful for shallow AR models in which a FIR filter
        is used for the target signals. For other types of model, you don't need to
        implement this method.

        Defaults to do nothing.

        Args:
            y (tensor): target features

        Returns:
            tensor: preprocessed target features
        """
        return y

    def prediction_type(self):
        """Prediction type.

        If your model has a MDN layer, please return ``PredictionType.PROBABILISTIC``.

        Returns:
            PredictionType: Determisitic or probabilistic. Default is deterministic.
        """
        return PredictionType.DETERMINISTIC

    def is_autoregressive(self):
        """Is autoregressive or not

        If your custom model is an autoregressive model, please return ``True``. In that case,
        you would need to implement autoregressive sampling in :py:meth:`inference`.

        Returns:
            bool: True if autoregressive. Default is False.
        """
        return False

    def has_residual_lf0_prediction(self):
        """Whether the model has residual log-F0 prediction or not.

        This should only be used for acoustic models.

        Returns:
            bool: True if the model has residual log-F0 prediction. Default is False.
        """
        return False



import pytest
import torch
from nnsvs.postfilters import Conv2dPostFilter, MovingAverage1d, MultistreamPostFilter
from nnsvs.util import init_seed


def _test_model_impl(model, in_dim):
    B = 4
    T = 100
    init_seed(B * T)
    x = torch.rand(B, T, in_dim)
    lengths = torch.Tensor([T] * B).long()

    # warmup forward pass
    with torch.no_grad():
        y = model(x, lengths)
        y_inf = model.inference(x, lengths)

    assert y.shape == (B, T, in_dim)
    assert y.shape == y_inf.shape


@pytest.mark.parametrize("noise_type", ["bin_wise", "frame_wise"])
def test_conv2d_postfilter(noise_type):
    params = {
        "in_dim": 60,
        "channels": 8,
        "kernel_size": (3, 3),
        "padding_mode": "zeros",
        "noise_type": noise_type,
        "init_type": "none",
        "smoothing_width": 5,
    }
    model = Conv2dPostFilter(**params)
    _test_model_impl(model, params["in_dim"])


@pytest.mark.parametrize("mgc_offset", [0, 1, 2])
@pytest.mark.parametrize("bap_offset", [0, 1])
def test_multistream_postfilter(mgc_offset, bap_offset):
    params = {
        "channels": 8,
        "kernel_size": (3, 3),
        "padding_mode": "zeros",
        "noise_type": "frame_wise",
        "init_type": "none",
    }
    mgc_postfilter = Conv2dPostFilter(**{**params, "in_dim": 60 - mgc_offset})
    bap_postfilter = Conv2dPostFilter(**{**params, "in_dim": 5 - bap_offset})

    # (mgc, lf0, vuv, bap)
    stream_sizes = [60, 1, 1, 5]

    model = MultistreamPostFilter(
        mgc_postfilter=mgc_postfilter,
        bap_postfilter=bap_postfilter,
        lf0_postfilter=None,
        stream_sizes=stream_sizes,
        mgc_offset=mgc_offset,
        bap_offset=bap_offset,
    )
    _test_model_impl(model, sum(stream_sizes))


@pytest.mark.parametrize("kernel_size", [1, 3, 5])
def test_moving_average_filter(kernel_size):
    T = 1000
    C = 4
    filt = MovingAverage1d(C, C, kernel_size)
    x = torch.randn(4, 1, T).expand(4, C, T)
    y = filt(x)
    assert x.shape == y.shape
    # make sure that the same filter is applied across channels
    for idx in range(C - 1):
        assert (y[0, idx, :] == y[0, idx + 1, :]).all()

    filt = MovingAverage1d(1, 1, kernel_size)
    x = torch.randn(4, 1, T)
    y = filt(x)
    assert x.shape == y.shape



import torch
from nnsvs.wavenet import WaveNet


def test_wavenet():
    x = torch.rand(16, 200, 206)
    c = torch.rand(16, 200, 300)

    model = WaveNet(in_dim=300, out_dim=206, layers=2)
    y = model(c, x)

    assert y.shape == x.shape

    model.eval()
    for T in [10, 20, x.shape[1]]:
        y = model.inference(c, num_time_steps=T)
        assert y.shape == (16, T, 206)



import re
from copy import deepcopy

import numpy as np
from nnmnkwii.io import hts

_flag_re = re.compile(r"\^([A-Za-z0-9]+)\_")


def full_to_mono(labels):
    """Convert full-context labels to mono labels

    Args:
        labels (nnmnkwii.io.hts.HTSLabelFile): HTS labels

    Returns:
        nnmnkwii.io.hts.HTSLabelFile: Mono HTS labels
    """
    is_full_context = "@" in labels.contexts[0]
    if not is_full_context:
        return labels

    mono_labels = deepcopy(labels)
    mono_labels.contexts = [c.split("-")[1].split("+")[0] for c in labels.contexts]

    return mono_labels


def get_note_frame_indices(binary_dict, numeric_dict, in_feats):
    """Get note frame indices from frame-level input features

    Note that the F0 in the input features must be discrete F0.

    Args:
        binary_dict (dict): Dictionary of binary features
        numeric_dict (dict): Dictionary of numeric features
        in_feats (np.ndarray): Input features

    Returns:
        np.ndarray: Note frame indices
    """
    pitch_idx = get_pitch_index(binary_dict, numeric_dict)
    score_f0 = in_feats[:, pitch_idx]
    note_frame_indices = np.where(score_f0 > 0)[0]
    return note_frame_indices


def get_pitch_index(binary_dict, numeric_dict):
    """Get pitch index from binary and numeric feature dictionaries

    Args:
        binary_dict (dict): Dictionary of binary features
        numeric_dict (dict): Dictionary of numeric features

    Returns:
        int: Pitch index
    """
    idx = 0
    pitch_idx = len(binary_dict)
    while idx < len(numeric_dict):
        if numeric_dict[idx][1].pattern.startswith("/E"):
            pitch_idx = pitch_idx + idx
            break
        idx += 1
    return pitch_idx


def get_pitch_indices(binary_dict, numeric_dict):
    """Get pitch indices from binary and numeric feature dictionaries

    Args:
        binary_dict (dict): Dictionary of binary features
        numeric_dict (dict): Dictionary of numeric features

    Returns:
        list: Pitch indices
    """
    idx = 0
    pitch_idx = len(binary_dict)
    assert np.any(
        [numeric_dict[idx][1].pattern.startswith(p) for p in ["/D", "/E", "/F"]]
    )
    pitch_indices = [pitch_idx]
    while True:
        idx += 1
        if np.any(
            [numeric_dict[idx][1].pattern.startswith(p) for p in ["/D", "/E", "/F"]]
        ):
            pitch_indices.append(pitch_idx + idx)
        else:
            break
    return pitch_indices


def get_note_indices(labels):
    """Get note start indices from HTS labels

    Args:
        labels (nnmnkwii.io.hts.HTSLabelFile): HTS labels

    Returns:
        list: Note start indices
    """
    note_indices = [0]
    last_start_time = labels.start_times[0]
    for idx in range(1, len(labels)):
        if labels.start_times[idx] != last_start_time:
            note_indices.append(idx)
            last_start_time = labels.start_times[idx]
        else:
            pass
    return note_indices


def merge_sil(labels):
    N = len(labels)
    f = hts.HTSLabelFile()
    f.append(labels[0], strict=False)
    is_full_context = "@" in labels[0][-1]
    for i in range(1, N):
        if (is_full_context and "-sil" in f[-1][-1] and "-sil" in labels[i][-1]) or (
            not is_full_context and f[-1][-1] == "sil" and labels[i][-1] == "sil"
        ):
            # extend sil
            f.end_times[-1] = labels[i][1]
        else:
            f.append(labels[i], strict=False)
    return f


def _is_br(label):
    is_full_context = "@" in label
    if is_full_context:
        is_br = "-br" in label
    else:
        is_br = label == "br"
    return is_br


def _is_silence(label):
    is_full_context = "@" in label
    if is_full_context:
        is_silence = "-sil" in label or "-pau" in label
    else:
        is_silence = label == "sil" or label == "pau"
    return is_silence


def compute_nosil_duration(labels, threshold=5.0):
    is_full_context = "@" in labels[0][-1]
    sum_d = 0
    for s, e, label in labels:
        d = (e - s) * 1e-7
        if is_full_context:
            is_silence = "-sil" in label or "-pau" in label
        else:
            is_silence = label == "sil" or label == "pau"
        if is_silence and d > threshold:
            pass
        else:
            sum_d += d
    return sum_d


def segment_labels(
    labels,
    strict=True,
    silence_threshold=0.1,
    min_duration=5.0,
    force_split_threshold=5.0,
):
    """Segment labels based on sil/pau

    Example:

    [a b c sil d e f pau g h i sil j k l]
    ->
    [a b c] [d e f] [g h i] [j k l]

    """
    seg = hts.HTSLabelFile()
    start_indices = []
    end_indices = []
    si = 0

    done_last_label = False
    for idx, (s, e, label) in enumerate(labels):
        d = (e - s) * 1e-7
        is_silence = _is_silence(label)

        if len(seg) > 0:
            # Compute duration except for long silences
            seg_d = compute_nosil_duration(seg)
        else:
            seg_d = 0

        # let's try to split
        if (is_silence and d > force_split_threshold) or (
            is_silence and d > silence_threshold and seg_d > min_duration
        ):
            if idx == len(labels) - 1:
                pass
            elif len(seg) > 0:
                start_indices.append(si)
                if is_silence and d > force_split_threshold:
                    end_indices.append(idx - 1)
                    start_indices.append(idx)
                    end_indices.append(idx)
                    seg = hts.HTSLabelFile()
                else:
                    seg.append((s, e, label), strict)
                    end_indices.append(idx)
                    seg = hts.HTSLabelFile()
                si = idx + 1
            elif len(seg) == 0:
                seg.append((s, e, label), strict)
                start_indices.append(si)
                end_indices.append(idx)
                seg = hts.HTSLabelFile()
        else:
            if len(seg) == 0:
                si = idx
            if idx == len(labels) - 1:
                done_last_label = True
            seg.append((s, e, label), strict)

    if len(seg) > 0:
        seg_d = compute_nosil_duration(seg)
        # If the last segment is short, combine with the previous segment.
        if seg_d < min_duration and len(end_indices) > 1:
            end_indices[-1] = si + len(seg) - 1
        else:
            start_indices.append(si)
            end_indices.append(si + len(seg) - 1)

        # Handle last label
        if not done_last_label:
            s, e, label = labels[-1]
            d = (e - s) * 1e-7
            if _is_silence(label) and d > silence_threshold:
                start_indices.append(end_indices[-1])
                end_indices.append(end_indices[-1])

    segments = []
    for s, e in zip(start_indices, end_indices):
        seg = labels[s : e + 1]

        offset = seg.start_times[0]
        seg.start_times = np.asarray(seg.start_times) - offset
        seg.end_times = np.asarray(seg.end_times) - offset

        segments.append(seg)

    return segments


def _label2phrases_neutrino(labels):
    """Segment HTS labels to phrases (NETRINO compatible)

    Args:
        labels (nnmnkwii.io.hts.HTSLabelFile): HTS labels

    Returns:
        list: phrases (i.e., list of HTS labels)
    """
    start_indices = []
    end_indices = []

    started = True
    start_indices.append(0)
    is_sil_phrase = _is_silence(labels.contexts[0])

    for idx, (_, _, label) in enumerate(labels):
        # sil or pau shouldn't be placed right before the br
        if idx > 0 and _is_br(label):
            assert not _is_silence(labels.contexts[idx - 1])

        # we are in the same phrase group
        if started:
            if is_sil_phrase:
                if _is_silence(label):
                    continue
            else:
                if (
                    not _is_silence(label)
                    and (idx > 0 and not _is_br(labels.contexts[idx - 1]))
                    or (idx == 0 and not _is_silence(label))
                ):
                    continue

        # reached the end of phrase
        end_indices.append(idx)
        started = False

        # start new phrase
        if not started:
            started = True
            is_sil_phrase = _is_silence(label)
            start_indices.append(idx)

    # handle last phrase
    if len(end_indices) == len(start_indices) - 1:
        end_indices.append(len(labels))

    # Make a list of HTS labels
    phrases = [labels[s:e] for (s, e) in zip(start_indices, end_indices)]
    return phrases, start_indices, end_indices


def fix_label_offset_to_zero(labels):
    """Fix label offset to zero

    Args:
        labels (nnmnkwii.io.hts.HTSLabelFile): HTS labels

    Returns:
        nnmnkwii.io.hts.HTSLabelFile: HTS labels with fixed offset
    """
    offset = labels.start_times[0]
    if offset > 0:
        labels.start_times = np.asarray(labels.start_times) - offset
        labels.end_times = np.asarray(labels.end_times) - offset
    return labels


def _label2phoneme_for_phrases(labels, s, e, note_indices=None):
    if s == e:
        r = labels.contexts[s]
    elif note_indices is None:
        r = " ".join(labels[s:e].contexts)
    else:
        rs = []
        for i in range(s, e):
            if i not in [s, e] and i in note_indices:
                rs.append(",")
            rs.append(labels.contexts[i])
        r = " ".join(rs).replace(" ,", ",")
    return r


def label2phrases_str(labels, note_indices):
    """Convert labels to NEUTRINO-format phraselist

    Note that timing labels should be used as input to get the same
    output as the NEUTRINO.

    Args:
        labels (nnmnkwii.io.hts.HTSLabelFile): HTS labels
        note_indices (list): indices of notes. This is needed to
            insert ``,`` at note boundaries.

    Returns:
        str: NEUTRINO-format phraselist
    """
    _, start_indices, end_indices = _label2phrases_neutrino(labels)

    out = ""
    for idx in range(len(end_indices)):
        s, e = start_indices[idx], end_indices[idx]
        start_time = int(labels.start_times[s] // 10000)
        ph = _label2phoneme_for_phrases(labels, s, e, note_indices)
        is_voiced_phrase = not ("sil" in ph or "pau" in ph)
        out += f"{idx} {start_time} {int(is_voiced_phrase)} {ph}\n"
    return out


def label2phrases(labels, fix_offset=True):
    """Convert labels to phrases

    The definision of a phrase is the same as NEUTRINO.
    See https://studio-neutrino.com/332/ for details.

    Args:
        labels (nnmnkwii.io.hts.HTSLabelFile): HTS labels
        fix_offset (bool): If True, fix label offset to zero

    Returns:
        list: phrases (i.e., list of HTS labels)
    """
    phrases = _label2phrases_neutrino(labels)[0]
    if fix_offset:
        phrases = [fix_label_offset_to_zero(p) for p in phrases]
    return phrases


def overwrite_phoneme_flags_(labels, flag):
    """Overwrite phoneme flags

    Args:
        labels (nnmnkwii.io.hts.HTSLabelFile): HTS labels
        flag (str): phoneme flag to overwrite

    Returns:
        nnmnkwii.io.hts.HTSLabelFile: modified HTS labels
    """

    contexts = labels.contexts
    for i in range(len(contexts)):
        n = len(_flag_re.findall(contexts[i]))
        if n == 0:
            print(i, contexts[i])
            print("Warn: it is likely to have a wrong input format. Ignoring.")
        elif n != 1:
            print(i, contexts[i])
            raise RuntimeError("More than two flags are found")
        contexts[i] = _flag_re.sub(f"^{flag}_", contexts[i])
    labels.contexts = contexts

    return labels
